"""
Dictionary utils

"""
from builtins import range
from future.utils import viewitems, viewkeys
import numpy
import pandas


def merge_dictionaries(dicts):
    """
    .. versionadded:: 0.3.1

    Merges keys and values from a list/iterable of dictionaries. The resulting
    dictionary's values are converted into sets, with the assumption that the
    values are one of the following: float, str, int, bool
    """
    merged = {}
    for d in dicts:
        for key, value in viewitems(d):
            if isinstance(value, (float, str, int, bool)):
                value = [value]
            try:
                merged[key].update(value)
            except KeyError:
                merged[key] = set(value)
    return merged


def combine_dict(keydict, valuedict):
    """
    Combine two dictionaries when the values of keydict are iterables. The
    combined dictionary has the same keys as keydict and the its values are
    sets containing all the values associated to keydict values in valuedict.

    .. digraph:: keydict
        :alt: key1 -> [v1, v2, .., vN]

        "keydict" -> "key1" -> "[v1, v2, .., vN]";

    .. digraph:: valuedict
        :alt: v1 -> [u1, u2, .., uN]
              v2 -> [t1, t2, .., tN]

        "valuedict" -> "v1" -> "[u1, u2, .., uN]";
        "valuedict" -> "v2" -> "[t1, t2, .., tN]";

    Resulting dictionary will be

    .. digraph:: combined
        :alt: key1->{u1, u2, .., uN}

        "combined" -> "key1" -> "{u1, u2, .., uN, t1, t2, .., tN}";

    :param dict keydict: dictionary whose keys are the same as the returned
        dictionary
    :param dict valuedict: dictionary whose values are the same as the returned
        dictionary

    :return dict: combined dictionary
    """

    comb_dict = dict((key, set()) for key in keydict)

    for key, values in keydict.items():
        for value in values:
            try:
                comb_dict[key].update(valuedict[value])
            except KeyError:
                # in case a value isn't in valuedict keys, silently pass
                pass

    return comb_dict


def combine_dict_one_value(keydict, valuedict):
    """
    Combine two dictionaries by the value of the keydict is used as a key in
    valuedict and the resulting dictionary is composed of keydict keys and
    valuedict values.

    Same as :func:`comb_dict`, but each value in keydict is a single element
    that is key in valuedict.

    :param dict keydict: dictionary whose keys are the same as the returned
        dictionary
    :param dict valuedict: dictionary whose values are the same as the returned
        dictionary

    :return dict: combined dictionary
    """

    comb_dict = {}

    for key, value in keydict.items():
        comb_dict[key] = valuedict[value]

    return comb_dict


def link_ids(id_map, black_list=None):
    """
    Given a dictionary whose values (iterables) can be linked back to other
    keys, it returns a dictionary in which the keys are the original keys and
    the values are sets of keys to which they can be linked.

    .. digraph:: input
        :alt: key1->[v1, v2]
              key2->[v3, v4]
              key3->[v2, v4]

        "id_map" -> "key1" -> "[v1, v2]";
        "id_map" -> "key2" -> "[v3, v4]";
        "id_map" -> "key3" -> "[v2, v4]";

    Becomes:

    .. digraph:: output
        :alt: key1->[key1, key3]
              key2->[key3]
              key3->[key1, key2]

        "linked" -> "key1" -> "[key1, key3]"
        "linked" -> "key2" -> "[key3]"
        "linked" -> "key3" -> "[key1, key2]"

    :param dict id_map: dictionary of keys to link
    :param iterable black_list: iterable of values to skip in making the links

    :return dict: linked dictionary
    """
    id_to_id = {}

    for s_id, s_cps in viewitems(id_map):
        id_links = set()
        for e_id, e_cps in viewitems(id_map):
            if e_id == s_id:
                continue
            for s_id2 in s_cps:
                if black_list is not None:
                    if s_id2 in black_list:
                        continue
                if s_id2 in e_cps:
                    id_links.add(e_id)
                    break
        id_to_id[s_id] = id_links
    return id_to_id


def reverse_mapping(map_dict):
    """
    Given a dictionary in the form: key->[v1, v2, .., vN], returns a dictionary
    in the form: v1->[key1, key2, .., keyN]

    :param dict map_dict: dictionary to reverse

    :return dict: reversed dictionary
    """
    rev_map = {}

    for key, value_ids in viewitems(map_dict):
        for value_id in value_ids:
            try:
                rev_map[value_id].add(key)
            except KeyError:
                rev_map[value_id] = set([key])

    return rev_map


def find_id_in_dict(s_id, s_dict):
    """
    Finds a value 's_id' in a dictionary in which the values are iterables.
    Returns a list of keys that contain the value.

    :param dict s_id: element to look for in the dictionary's values
    :param object d: dictionary to search in

    :return list: list of keys in which d was found
    """
    f_list = []
    for k_id, v_ids in s_dict.items():
        if s_id in v_ids:
            f_list.append(k_id)
    return f_list


def split_dictionary_by_value(value_dict, threshold, aggr_func=numpy.median,
                              key_filter=None):
    """
    Splits a dictionary, whose values are iterables, based on a threshold:

        * one in which the result of aggr_func is lower than the threshold
          (first)
        * one in which the result of aggr_func is equal or greater than the
          threshold (second)

    :param dict valuedict: dictionary to be splitted
    :param number threshold: must be comparable to threshold
    :param func aggr_func: function used to aggregate the dictionary values
    :param iterable key_filter: if specified, only these key will be in the
        resulting dictionary

    :return: two dictionaries
    """
    lower_dict = {}
    higher_dict = {}

    if key_filter is None:
        key_filter = viewkeys(value_dict)

    for key in key_filter:
        values = value_dict[key]
        if aggr_func(values) < threshold:
            lower_dict[key] = values
        else:
            higher_dict[key] = values

    return lower_dict, higher_dict


def apply_func_to_values(dictionary, func):
    """
    .. versionadded:: 0.1.12

    Assuming a dictionary whose values are iterables, *func* is applied to each
    element of the iterable, retuning a *set* of all transformed elements.

    Arguments:
        dictionary (dict): dictionary whose values are iterables
        func (func): function to apply to the dictionary values

    Returns:
        dict: dictionary with transformed values
    """
    return dict(
        (key, set(func(value) for value in values))
        for key, values in viewitems(dictionary)
    )


def filter_ratios_by_numbers(ratios, min_num):
    """
    Returns from a dictionary only the items for which the length of the
    iterables that is the value of the item, is equal or greater of min_num.

    :param dict ratios: dictionary key->list
    :param int min_num: minimum number of elements in the value iterable

    :return dict: filtered dictionary
    """
    return dict(
        (key, values) for key, values in viewitems(ratios)
        if len(values) >= min_num
    )


def filter_nan(ratios):
    """
    Returns a dictionary with the NaN values taken out
    """

    return dict(
        (key, [ratio for ratio in ratios[key] if not numpy.isnan(ratio)])
        for key in ratios
    )


class cache_dict_file(object):
    """
    .. versionadded:: 0.3.0

    Used to cache the result of a function that yields a tuple (key and value).
    If the value is found in the internal dictionary (as the class behave), the
    correspondent value is returned, otherwise the iterator is advanced until
    the key is found.

    Example:
        >>> from mgkit.io.blast import parse_accession_taxa_table
        >>> i = parse_accession_taxa_table('nucl_gb.accession2taxid.gz', key=0)
        >>> d = cache_dict_file(i)
        >>> d['AH001684']
        4400
    """
    _iterator = None
    _dict = None

    def __init__(self, iterator, skip_lines=0):
        """
        Arguments:
            iterator (iter): iterator used in building the dictionary
            skip_lines (int): how many iterations to skip at the start
        """
        for index in range(skip_lines):
            next(iterator)
        self._iterator = iterator
        self._dict = {}

    def __getitem__(self, key):
        try:
            value = self._dict[key]
        except KeyError:
            while True:
                nkey, nvalue = self.next()
                if nkey == key:
                    value = nvalue
                    break
        return value

    def next(self):
        try:
            key, value = next(self._iterator)
        except StopIteration:
            raise KeyError
        self._dict[key] = value
        return key, value


class HDFDict(object):
    """
    .. versionchanged:: 0.3.3
        added *cache* in __init__

    .. versionadded:: 0.3.1

    Used a table in a HDFStore (from pandas) as a dictionary. The table must be
    indexed to perform well. Read only.

    .. note::

        the dictionary cannot be modified and exception:`ValueError` will be
        raised if the table is not in the file
    """
    def __init__(self, file_name, table, cast=int, cache=True):
        self._hdf = pandas.HDFStore(file_name, mode='r')
        self._table = table
        self._cast = cast
        if self._table not in self._hdf:
            raise ValueError(
                "Table ({}) not found in file ({})".format(
                    table,
                    file_name
                )
            )
            self._hdf.close()

        if cache:
            self.__getitem__ = self._getitem_cache
            self._cache = {}
        else:
            self.__getitem__ = self._getitem_hdf

    def _getitem_cache(self, key):
        value = self._cache.get(key, None)
        if value is None:
            value = self._getitem_hdf(key)
            self._cache[key] = value
            return value

    def _getitem_hdf(self, key):
        df = self._hdf.select(self._table, 'index=key')
        if df.empty:
            raise KeyError('Key not found {}'.format(key))
        return self._cast(df.values)

    __getitem__ = _getitem_hdf


def dict_to_text(stream, dictionary, header=None, comment=None, sep='\t'):
    """
    .. versionadded:: 0.4.4

    Writes the content of a dictionary to a stream (supports *write*), like
    io.StringIO or an opened file. Intended to be used only for dictionaries
    with key-value of type integer/strings, other data types are better served
    by more complex options, like JSON, etc.

    .. warning::

        The file is expected to be opened in text mode ('r')

    Arguments:
        stream (file): stream to write to, to output a string, use io.StringIO
        dictionary (dict): dictionary to write
        header (iterable): a tuple/list to be used as header
        comment (str): a comment at the start of the file - '# ' will be
            prepended to the value passed.
        sep (str): column separator to use
    """
    if comment is not None:
        stream.write('# {}\n'.format(comment))
    if header is not None:
        stream.write('{}\n'.format(sep.join(header)))
    for key, value in dictionary.items():
        stream.write(f'{key}{sep}{value}\n')


def text_to_dict(stream, skip_lines=0, sep='\t', key_index=0, value_index=1,
                 key_func=str, value_func=str, encoding=None):
    """
    .. versionadded:: 0.4.4

    Reads a dictionary form a table file, the passed file is assumed to be
    opened as text, not binary - in which case you need to pass the encoding
    (e.g. *ascii*). The file may have multiple columns, so the key and value
    columns can be chosen with *key_index* and *value_index*, respectively.

    Arguments:
        stream (file): stream that can be read as a file
        skip_lines (int): number of lines to skip at the start of the file
        sep (str): column separator to use
        key_index (int): zero-based column number of keys
        value_index (int): zero-based column number of values
        key_func (func): function to apply to the keys (defaults to *str*)
        value_func (func): function to apply to the values (defaults to *str*)
        encoding (None, str): if *None* is passed, the file is assumed to be
            opened in text mode, otherwise the encoding of the file must be
            passed

    Yields:
        tuple: the keys and values that can be passed to *dict*
    """

    for index, line in enumerate(stream):

        if skip_lines > 0:
            skip_lines -= 1
            continue

        if encoding is not None:
            line = line.decode(encoding)

        line = line.rstrip()

        # line empty, so skip
        if not line:
            continue

        line = line.split(sep)

        key = key_func(line[key_index])
        value = value_func(line[value_index])

        yield key, value



"""
Module containing classes and functions to deal with eggNOG data

.. todo::

    * unify download of data from web

"""
from __future__ import print_function
from future.utils import viewitems
import logging
import itertools
from ..io import open_file
from ..utils import dictionary

LOG = logging.getLogger(__name__)

EGGNOG_CAT_NAMES = (
    'INFORMATION STORAGE AND PROCESSING',
    'CELLULAR PROCESSES AND SIGNALING',
    'METABOLISM',
    'POORLY CHARACTERIZED'
)
"""
Functional categories (broader)
"""

EGGNOG_CAT_KEYS = (
    ('J', 'A', 'K', 'L', 'B'),
    ('D', 'Y', 'V', 'T', 'M', 'N', 'Z', 'W', 'U', 'O'),
    ('C', 'G', 'E', 'F', 'H', 'I', 'P', 'Q'),
    ('R', 'S')
)
"""
Used to build map of broader categories (:data:`EGGNOG_CAT_NAMES`) to more
specific ones
"""

EGGNOG_CAT_MAP = dict(
    (label, categories)
    for label, categories in zip(EGGNOG_CAT_NAMES, EGGNOG_CAT_KEYS)
)
"""
Functional categories (broader, :data:`EGGNOG_CAT_NAMES`) mappings to more
specific one (:data:`EGGNOG_CAT`).
"""

EGGNOG_CAT = {

    # INFORMATION STORAGE AND PROCESSING
    'J': "Translation, ribosomal structure and biogenesis",
    'A': "RNA processing and modification",
    'K': "Transcription",
    'L': "Replication, recombination and repair",
    'B': "Chromatin structure and dynamics",

    # CELLULAR PROCESSES AND SIGNALING
    'D': "Cell cycle control, cell division, chromosome partitioning",
    'Y': "Nuclear structure",
    'V': "Defense mechanisms",
    'T': "Signal transduction mechanisms",
    'M': "Cell wall/membrane/envelope biogenesis",
    'N': "Cell motility",
    'Z': "Cytoskeleton",
    'W': "Extracellular structures",
    'U': "Intracellular trafficking, secretion, and vesicular transport",
    'O': "Posttranslational modification, protein turnover, chaperones",

    # METABOLISM
    'C': "Energy production and conversion",
    'G': "Carbohydrate transport and metabolism",
    'E': "Amino acid transport and metabolism",
    'F': "Nucleotide transport and metabolism",
    'H': "Coenzyme transport and metabolism",
    'I': "Lipid transport and metabolism",
    'P': "Inorganic ion transport and metabolism",
    'Q': "Secondary metabolites biosynthesis, transport and catabolism",

    # POORLY CHARACTERIZED
    'R': "General function prediction only",
    'S': "Function unknown"
}
"""
Single letter functional categories
"""


def get_general_eggnog_cat(category):
    """
    .. versionadded:: 0.1.14

    Returns the functional category (:data:`EGGNOG_CAT_NAMES` keys)
    for the requested single letter functional category (:data:`EGGNOG_CAT`
    keys)
    """
    return set(
        gen_category
        for gen_category, categories in viewitems(EGGNOG_CAT_MAP)
        if category in categories
    )


class NOGInfo(object):
    """
    .. versionadded:: 0.1.14

    .. versionchanged:: 0.4.0
        made file reading compatible with Python 3

    Mappings from Uniprot to eggNOG

    ..note::

        load_description is optional
    """
    _map_nog_func = None
    "eggNOG COG/NOG to functional category dictionary"
    _map_nog_gene = None
    "eggNOG COG/NOG to gene id dictionary"
    _map_nog_desc = None
    "eggNOG id to description dictionary"
    _map_gene_nog = None
    "eggNOG gene id to COG/NOG dictionary"

    def __init__(self, members=None, funccat=None, description=None):
        """
        .. versionchanged:: 0.4.0
            added parameters to load data at __init__
        """
        self._map_nog_gene = {}
        self._map_gene_nog = {}
        self._map_nog_func = {}
        self._map_nog_desc = {}
        if members is not None:
            self.load_members(members)
        if funccat is not None:
            self.load_funccat(funccat)
        if description is not None:
            self.load_description(description)

    def load_members(self, file_handle):
        """
        Loads data from *NOG.members.txt.gz*

        *file_handle* can either an open file or a path
        """
        file_handle = open_file(file_handle, 'rb')

        LOG.info("Reading Members from %s", file_handle.name)

        map_nog_gene = {}

        for line in file_handle:
            line = line.decode('ascii')
            if line.startswith('#'):
                continue

            nog_id, gene_id, start, end = line.strip().split()

            try:
                map_nog_gene[nog_id].append(gene_id)
            except KeyError:
                map_nog_gene[nog_id] = [gene_id]

        self._map_nog_gene.update(map_nog_gene)
        self._map_gene_nog.update(dictionary.reverse_mapping(map_nog_gene))

    def load_description(self, file_handle):
        """
        Loads data from *NOG.description.txt.gz*

        *file_handle* can either an open file or a path
        """
        file_handle = open_file(file_handle, 'rb')

        LOG.info("Reading Descriptions from %s", file_handle.name)

        map_nog_desc = {}

        for line in file_handle:
            line = line.decode('ascii')
            try:
                nog_id, nog_desc = line.strip().split('\t', 1)
            except ValueError:
                nog_id = line.strip()
                nog_desc = ''

            map_nog_desc[nog_id] = nog_desc

        self._map_nog_desc.update(map_nog_desc)

    def load_funccat(self, file_handle):
        """
        Loads data from *NOG.funccat.txt.gz*

        *file_handle* can either an open file or a path
        """
        file_handle = open_file(file_handle, 'rb')

        LOG.info("Reading Functional Categories from %s", file_handle.name)

        map_nog_func = {}

        for line in file_handle:
            line = line.decode('ascii')
            nog_id, nog_func = line.strip().split()
            nog_func = set(nog_func)

            map_nog_func[nog_id] = nog_func

        self._map_nog_func.update(map_nog_func)

    def get_nog_funccat(self, nog_id):
        """
        Returns the functional category (one letter, :data:`EGGNOG_CAT` keys)
        for the requested eggNOG COG/NOG ID
        """
        try:
            return self._map_nog_func[nog_id].copy()
        except KeyError:
            return set()

    def get_nogs_funccat(self, nog_ids):
        """
        Returns the functional categories for a list of COG/NOG IDs. Uses
        :meth:`NOGInfo.get_nog_funccat`
        """

        iterator = (self.get_nog_funccat(nog_id) for nog_id in nog_ids)

        return set(itertools.chain(*iterator))

    def get_nog_gencat(self, nog_id):
        """
        Returns the functional category (:data:`EGGNOG_CAT_NAMES` keys)
        for the requested eggNOG COG/NOG IDs
        """
        return set(
            itertools.chain(
                *(
                    get_general_eggnog_cat(funccat)
                    for funccat in self.get_nog_funccat(nog_id)
                )
            )
        )

    def get_gene_nog(self, gene_id):
        """
        Returns the COG/NOG ID of the requested eggNOG gene ID
        """
        try:
            nog_ids = self._map_gene_nog[gene_id]
        except KeyError:
            return None

        return nog_ids

    def get_gene_funccat(self, gene_id):
        """
        Returns the functional category (one letter, :data:`EGGNOG_CAT` keys)
        for the requested eggNOG gene ID
        """
        try:
            nog_ids = self._map_gene_nog[gene_id]
        except KeyError:
            return None

        return self.get_nogs_funccat(nog_ids)



import pytest
from conftest import skip_no_connection
import random
from mgkit.mappings.eggnog import NOGInfo

pytestmark = skip_no_connection


@pytest.fixture(scope='module')
def nog_info(eggnog_v3):
    eg = NOGInfo(**eggnog_v3)
    return eg


def test_load_members(eggnog_v3):
    eg = NOGInfo()
    eg.load_members(eggnog_v3['members'])
    assert eg._map_gene_nog
    assert eg._map_nog_gene


def test_load_description(eggnog_v3):
    eg = NOGInfo()
    eg.load_description(eggnog_v3['description'])
    assert eg._map_nog_desc


def test_load_funccat(eggnog_v3):
    eg = NOGInfo()
    eg.load_funccat(eggnog_v3['funccat'])
    assert eg._map_nog_func


def test_noginfo_get_nog_funccat1(nog_info):
    nog_id = 'NOG19649'
    assert nog_info.get_nog_funccat(nog_id)


def test_noginfo_get_nog_funccat2(nog_info):
    assert nog_info.get_nog_funccat('XXXXXXX') == set()


def test_noginfo_get_nogs_funccat1(nog_info):
    assert nog_info.get_nogs_funccat(
        ['NOG19649', 'NOG134552', 'NOG74662']
    ) == set(['S'])


def test_noginfo_get_nog_gencat(nog_info):
    nog_id = 'NOG74662'
    assert nog_info.get_nog_gencat(nog_id) == {'POORLY CHARACTERIZED'}


def test_noginfo_gene_nog1(nog_info):
    assert nog_info.get_gene_nog('NOG19649') is None


def test_noginfo_gene_nog2(nog_info):
    assert nog_info.get_gene_nog('5062.CADAORAP00008010') == {'NOG129254'}


def test_noginfo_get_gene_funccat1(nog_info):
    assert nog_info.get_gene_funccat('5062.CADAORAP00008010') == set()


def test_noginfo_get_gene_funccat2(nog_info):
    assert nog_info.get_gene_funccat('177437.HRM2_09030') == {'K'}


def test_noginfo_get_gene_funccat3(nog_info):
    assert nog_info.get_gene_funccat('XXXXXX') is None



"""
Module dealing with BAM/SAM files
"""
from future.utils import viewitems
import logging
try:
    # In Python2
    from itertools import izip as zip
except ImportError:
    pass
from builtins import object
import pandas
import numpy
import pysam
from tqdm import tqdm
from mgkit.io.utils import open_file

LOG = logging.getLogger(__name__)


def get_region_coverage(bam_file, seq_id, feat_from, feat_to):
    """
    Return coverage for an annotation.

    .. note::

        feat_from and feat_to are 1-based indexes

    :param Samfile bam_file: instance of :class:`pysam.Samfile`
    :param str seq_id: sequence id
    :param int feat_from: start position of feature
    :param int feat_to: end position of feature

    :return int: coverage array for the annotation
    """
    norm_start, norm_end = feat_from - 1, feat_to - 1
    iterator = bam_file.pileup(
        reference=seq_id,
        start=norm_start,
        end=norm_end,
        truncate=True
    )

    coverage = pandas.Series(
        dict(
            (pileup_proxy.pos, pileup_proxy.n)
            for pileup_proxy in iterator
        ),
        dtype=numpy.int,
    )

    return coverage


def covered_annotation_bp(files, annotations, min_cov=1, progress=False):
    """
    .. versionadded:: 0.1.14

    Returns the number of base pairs covered of annotations over multiple
    samples.

    Arguments:
        files (iterable): an iterable that returns the alignment file names
        annotations (iterable): an iterable that returns annotations
        min_cov (int): minumum coverage for a base to counted
        progress (bool): if *True*, a progress bar is used

    Returns:
        dict: a dictionary whose keys are the uid and the values the number of
        bases that are covered by reads among all samples

    """
    annotations = [
        (annotation.uid, annotation.seq_id, annotation.start, annotation.end)
        for annotation in annotations
    ]

    covered = {}

    for file_name in files:
        # pysam 0.8.1 changed Samfile to AlignmentFile
        alg_file = pysam.AlignmentFile(file_name, 'rb')

        if progress:
            annotations = tqdm(annotations)

        for uid, seq_id, start, end in annotations:
            cov = get_region_coverage(alg_file, seq_id, start, end)

            try:
                covered[uid] = cov.add(covered[uid], fill_value=0)
            except KeyError:
                covered[uid] = cov

    return dict(
        (uid, len(cov[cov >= min_cov])) for uid, cov in viewitems(covered)
    )


def add_coverage_info(annotations, bam_files, samples, attr_suff='_cov'):
    """
    .. versionchanged:: 0.3.4
        the coverage now is returned as floats instead of int

    Adds coverage information to annotations, using BAM files.

    The coverage information is added for each sample as a 'sample_cov' and the
    total coverage as as 'cov' attribute in the annotations.

    .. note::

        The bam_files and sample variables must have the same order

    :param iterable annotations: iterable of annotations
    :param iterable bam_files: iterable of :class:`pysam.Samfile` instances
    :param iterable sample: names of the samples for the BAM files
    """
    LOG.info("Adding coverage info for %d annotations", len(annotations))
    for bam_file, sample in zip(bam_files, samples):
        LOG.info("Sample %s, file %s", sample, bam_file.filename)

    tot_coverage = {}

    for bam_file, sample in zip(bam_files, samples):
        LOG.info("Adding coverage for sample %s", sample)
        for index, annotation in enumerate(annotations):
            sample_coverage = get_region_coverage(
                bam_file,
                annotation.seq_id,
                annotation.start,
                annotation.end
            )
            # adds the results of the coverage to the total coverage
            # uses the add method of a Series to make sure that possible
            # nan values are filled with 0 before the sum
            try:
                tot_coverage[annotation.uid] = sample_coverage.add(
                    tot_coverage[annotation.uid],
                    fill_value=0
                )
            except KeyError:
                tot_coverage[annotation.uid] = sample_coverage

            cov_mean = sample_coverage.mean()

            annotation.set_attr(
                sample + attr_suff,
                0 if numpy.isnan(cov_mean) else int(cov_mean)
            )
            if (index + 1) % 1000 == 0:
                LOG.debug(
                    'Added coverage for (%d/%d) annotations',
                    index + 1,
                    len(annotations)
                )
    for annotation in annotations:
        cov_mean = tot_coverage[annotation.uid].mean()
        annotation.set_attr(
            'cov',
            0 if numpy.isnan(cov_mean) else cov_mean
        )


def read_samtools_depth(file_handle, num_seqs=10000, seq_ids=None):
    """
    .. versionchanged:: 0.4.2
        the function returns **lists** instead of numpy arrays for speed (at
        least in my tests it seems ~4x increase)

    .. versionchanged:: 0.4.0
        now returns 3 array, instead of 2. Also added *seq_ids* to skip lines

    .. versionchanged:: 0.3.4
        *num_seqs* can be None to avoid a log message

    .. versionadded:: 0.3.0

    Reads a samtools *depth* file, returning a generator that yields the
    array of each base coverage on a per-sequence base.

    .. note::

        There's no need anymore to use `samtools depth -aa`, because the
        function returns the position array and this can be used to create a
        Pandas SparseArray which can be reindexed to include missing positions
        (with values of 0)

        **Valid for version < 0.4.0**:

        The information on position is not used, to use numpy and save memory.
        samtools *depth* should be called with the `-aa` option::

             `samtools depth -aa bamfile`

        This options will output both base position with 0 coverage and
        sequneces with no aligned reads

    Arguments:
        file_handle (file): file handle of the coverage file
        num_seqs (int or None): number of sequence that fires a log message. If
            None, no message is triggered
        seq_ids (dict, set): a hashed container like a dictionary or set with
            the sequences to return

    Yields:
        tuple: the first element is the sequence identifier, the second one
        is the list with the positions, the third element is the list with the
        coverages
    """
    curr_key = ''
    curr_pos = []
    curr_cov = []

    file_handle = open_file(file_handle, 'rb')

    LOG.info(
        'Reading coverage from file (%s)',
        getattr(file_handle, 'name', repr(file_handle))
    )
    line_no = 0
    for line in file_handle:
        line = line.decode('ascii')
        # From Python3 the default is Universal newlines, and it's not expected
        # to have more than '\n' at the end of the line - increases speed
        # slightly
        name, pos, cov = line[:-1].split('\t')
        if (seq_ids is not None) and (name not in seq_ids):
            continue
        # only converts if sequence is to be used
        pos = int(pos)
        cov = int(cov)

        if curr_key == name:
                curr_pos.append(pos)
                curr_cov.append(cov)
        else:
            if curr_key == '':
                curr_cov.append(cov)
                curr_pos.append(pos)
                curr_key = name
            else:
                line_no += 1
                if (num_seqs is not None) and (line_no % num_seqs == 0):
                    LOG.info('Read %d sequence coverage', line_no)
                yield curr_key, curr_pos, curr_cov
                curr_key = name
                curr_cov = [cov]
                curr_cov = [pos]
    else:
        yield curr_key, curr_pos, curr_cov

    LOG.info('Read a total of %d sequence coverage', line_no + 1)


class SamtoolsDepth(object):
    """
    .. versionchanged:: 0.4.2
        several optimisations and changes to support a scanning approach,
        instead of lookup table. No exception is raised when a sequence is not
        found in the file, instead assuming that the coverage is 0

    .. versionchanged:: 0.4.0
        uses pandas.SparseArray now. It should use less memory, but needs
        pandas version > 0.24

    .. versionadded:: 0.3.0

    This a class that helps use the results from :func:`read_samtools_depth`.

    There are 2 modes of operations:

        1) Request a region coverage via :meth:`SamtoolsDepth.region_coverage`

        2) Advance the `samtools depth` file until a sequence coverage is read

    The method 1) was the default in MGKit < 0.4.2, when the user would request
    a region coverage and this class would advance the reading until the
    requested region is found. While scanning, all the sequences encountered
    are kept as pandas Series with SparseArray declared, reindexed from 0 to
    the *max_size_dict* values if provided, or *max_size*. The advantage is that
    the it's easier to use, but with bigger datasets will keep high memory
    usage. This is the case for case 2) which involves calling directly
    :meth:`SamtoolsDepth.advance_file` returning the sequence ID read. Then the
    region coverage can be requested the same way as before, but won't involve
    scanning the file.

    The use of a SparseArray in a pandas Series allows the use of
    `samtools depth` files that weren't produced with `samtools depth -aa` and
    save memory.

    .. note::

        The amount of memory used is fairly high, especially with the increasing
        number of sequences in a GFF/Depth file. It is recoommended to use
        max_size_dict to 1) only creates arrays for sequences needed and 2) use
        arrays of smaller size

    .. warning::

        In MGKit 0.4.2, the internal dictionary to keep the SparseArray(s)
        is a :class:`weakref.WeakValueDictionary`, to improve the release
        of memory. That on Linux seems to release the array **too fast**.
        Upgrade to MGKit version 0.4.3 if this class is needed

    """
    file_handle = None
    data = None
    max_size = None
    max_size_dict = None
    not_found = None
    closed = False
    density = None

    def __init__(self, file_handle, num_seqs=10**4, max_size=10**6,
                 max_size_dict=None, calc_density=False, dtype='uint32'):
        """
        .. versionchanged:: 0.4.2
            added *raise_error* to control sequences not found in depth files,
            *calc_density* to debug the density of the SparseArray used and
            also dtype to control the dtype to use in the SparseArray

        .. versionchanged:: 0.4.0
            added *max_size* and max_size_dict

        Arguments:
            file_handle (file): the file handle to pass to
                :func:`read_samtools_depth`
            num_seqs (int): number of sequence that fires a log message
            max_size (int): max size to use in the SparseArray
            max_size_dict (dict): dictionary with max size for each seq_id
                requested. If None, *max_size* is used
            calc_density (bool): If True, the density of the SparseArray(s)
                used are kept in :attr:`SamtoolsDepth.density`
            dtype (str): dtype to pass to the SparseArray

        """
        self.max_size = max_size
        if max_size_dict is None:
            self.max_size_dict = {}
        else:
            self.max_size_dict = max_size_dict

        self.file_handle = read_samtools_depth(
            file_handle,
            num_seqs=num_seqs,
            seq_ids=max_size_dict
        )
        self.not_found = set()
        self.closed = False
        self.data = {}
        self.dtype = 'Sparse[{}]'.format(dtype)
        if calc_density:
            self.density = []

    def advance_file(self):
        """
        .. versionadded:: 0.4.2

        Requests the scan of the file and returns the sequence found, the
        coverage can then be retrivied with :meth:`SamtoolsDepth.region_coverage`

        Returns:
            (str, None): the return value is None if the file is fully read,
            otherwise the the sequnece ID found is returned
        """
        # continue scanning the file
        try:
            key, pos_array, cov_array = next(self.file_handle)
        except StopIteration:
            self.closed = True
            return None
        # Init a SparseArray (by passing the correct dtype)
        value = pandas.Series(
            # brings start position to 0
            dict(zip((pos - 1 for pos in pos_array), cov_array)),
            dtype=self.dtype
        # reindex to the correct size to allow
        ).reindex(
            range(self.max_size_dict.get(key, self.max_size)),
            fill_value=0
        )
        self.data[key] = value
        if self.density is not None:
            self.density.append(value.sparse.density)
        return key

    def region_coverage(self, seq_id, start, end):
        """
        .. versionchanged:: 0.4.2
            now using :meth:`SamtoolsDepth.advance_file` to scan the file

        Returns the mean coverage of a region. The *start* and *end* parameters
        are expected to be 1-based coordinates, like the correspondent
        attributes in :class:`mgkit.io.gff.Annotation` or
        :class:`mgkit.io.gff.GenomicRange`.

        If the sequence for which the coverage is requested is not found, the
        *depth* file is read (and cached) until it is found.

        Arguments:
            seq_id (str): sequence for which to return mean coverage
            start (int): start of the region
            end (int): end of the region

        Returns:
            float: mean coverage of the requested region
        """
        if seq_id in self.data:
            # data already in dictionary
            return self.data[seq_id][start-1:end].mean()
        elif seq_id in self.not_found:
            # already asked for and not found in depth file
            return 0.
            # depth file is closed and we assume
            # so no coverage is available
        elif self.closed:
            self.not_found.add(seq_id)
            return 0.
        else:
            while True:
                key = self.advance_file()
                if key is None:
                    self.not_found.add(seq_id)
                    self.closed = True
                    return 0.
                elif key == seq_id:
                    return self.data[seq_id][start-1:end].mean()

    def drop_sequence(self, seq_id):
        """
        .. versionadded:: 0.4.2

        Remove the sequence passed from the internal dictionary
        """
        if seq_id in self.data:
            del self.data[seq_id]



"Module configuring log information"

import click
import logging
import sys
import os.path

DEBUG_FMT = "%(asctime)s - %(levelname) 7s - %(name)s->%(funcName)s: " + \
    "%(message)s"
INFO_FMT = "%(levelname)s - %(name)s: %(message)s"


class ColorFormatter(logging.Formatter):
    colors = {
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'magenta',
        'CRITICAL': 'red',
        'DEBUG': 'blue',
    }
    
    def format(self, record):
        color = self.colors.get(record.levelname, None)
        record.levelname = click.style(record.levelname, fg=color)
        return super(ColorFormatter, self).format(record)


def config_log(level=logging.DEBUG, output=sys.stderr):
    """
    Minimal configuration of :mod`logging` module, default to debug level and
    the output is printed to standard error

    :param int level: logging level
    :param file output: file to which write the log
    """
    log_handler = logging.StreamHandler(output)

    if level == logging.DEBUG:
        fmt = ColorFormatter(fmt=DEBUG_FMT)
    else:
        fmt = ColorFormatter(fmt=INFO_FMT)

    log_handler.setFormatter(fmt)
    log_handler.setLevel(level)

    logging.getLogger().addHandler(log_handler)
    logging.getLogger().setLevel(level)


def config_log_to_file(level=logging.DEBUG, output=None):
    """
    .. versionadded:: 0.1.14

    Minimal configuration of :mod`logging` module, default to debug level and
    the output is printed to script name, using `sys.argv[0]`.

    :param int level: logging level
    :param file output: file to which write the log
    """

    if output is None:
        output = os.path.join(
            os.path.splitext(sys.argv[0])[0],
            '.log'
        )

    log_handler = logging.FileHandler(filename=output, mode='w', delay=True)

    if level == logging.DEBUG:
        fmt = logging.Formatter(fmt=DEBUG_FMT)
    else:
        fmt = logging.Formatter(fmt=INFO_FMT)

    log_handler.setLevel(level)
    log_handler.setFormatter(fmt)

    logging.getLogger().addHandler(log_handler)
    logging.getLogger().setLevel(level)



import pytest
from mgkit.io import fasta
from mgkit.io.utils import open_file
import pathlib

@pytest.fixture
def nucseq(shared_datadir):
    return str(shared_datadir / 'test-seq-nuc.fa')


@pytest.fixture
def oneseq(shared_datadir):
    return str(shared_datadir / 'one-seq.fa')


def test_load_fasta1(nucseq):
    seq_id, seq = next(fasta.load_fasta(nucseq))
    assert seq == u'ATCGATGTCTGCCGCAATGACGGTGGCACCACGCTGGGCAGAGCGTACGACGGCACCCATACCGACCATGCCACAGCCAATCACCATGACGACATCGATATCTGTTACCTGGGCACGACTGACAGCATGGAAACCCACACTCATCGGTTCGATTAATGCACAGGTGCGGGGTGTCAGCAGTCCTGCGGGGATGACTTTCTCCCAGGGCAGGGCGAGATACTCACACATGGCTCCCCAGCGCTGCACACCTAATGTCTGGTTGTGCTCGCAGGCATTGACACGGTCGTTACGGCATGACGCACATTTTCCACAGTTGGTGTAGGGGTTGACGGTGACGGTCATACCGGGCTTCAGTCCCTCAGGTACGTTCTTGCCAATCTTGACAATCTCCGCACCTACCTCATGACCGGGAACCACAGGCATCTTCACCATCGGGTTACCGCCACGGAAGGTATTCAGGTCACTGCCGCAAAAACC'


def test_load_fasta2(nucseq):
    seq_id, seq = next(fasta.load_fasta(nucseq))
    assert seq_id == u'contig-1467318'


def test_load_fasta3(nucseq):
    count = sum(1 for x in fasta.load_fasta(nucseq))
    assert count == 115


def test_load_fasta4(oneseq):
    count = sum(1 for x in fasta.load_fasta(oneseq))
    assert count == 1


def test_load_fasta5(oneseq):
    seq_id, seq = next(fasta.load_fasta(oneseq))
    assert len(seq) == 630


def test_write_fasta_sequence1(nucseq, tmpdir):

    seq_id, seq = next(fasta.load_fasta(nucseq))
    file_name = (tmpdir / 'test.fa').strpath
    file_handle = open_file(file_name, 'w')

    fasta.write_fasta_sequence(file_handle, seq_id, seq)
    file_handle.close()

    seq_idw, seqw = next(fasta.load_fasta(file_name))

    assert (seq_id, seq) == (seq_idw, seqw)


def test_write_fasta_sequence2(nucseq, tmpdir):

    file_name = (tmpdir / 'test.fa').strpath
    file_handle = open_file(file_name, 'w')

    for seq_id, seq in fasta.load_fasta(nucseq):
        fasta.write_fasta_sequence(file_handle, seq_id, seq)
    file_handle.close()

    count1 = sum(1 for x in fasta.load_fasta(nucseq))
    count2 = sum(1 for x in fasta.load_fasta(file_name))

    assert count1 == count2


def test_split_fasta_file1(nucseq, tmpdir):

    fasta.split_fasta_file(nucseq, (tmpdir / 'test{}.fa').strpath, 3)

    files = list(pathlib.Path(tmpdir.strpath).glob('*.fa'))

    assert len(files) == 3


def test_split_fasta_file2(nucseq, tmpdir):

    fasta.split_fasta_file(nucseq, (tmpdir / 'test{}.fa').strpath, 3)

    files = list(
        str(path)
        for path in pathlib.Path(tmpdir.strpath).glob('*.fa')
    )

    count1 = sum(1 for x in fasta.load_fasta(nucseq))
    count2 = sum(1 for x in fasta.load_fasta_files(files))

    assert count1 == count2


def test_load_fasta_rename1(oneseq):
    seq_id, seq = next(fasta.load_fasta_rename(oneseq))
    assert seq_id == 'M95099.1'


def test_load_fasta_rename2(oneseq):
    nam_func = lambda x: x.split('.')[0]
    seq_id, seq = next(fasta.load_fasta_rename(oneseq, name_func=nam_func))
    assert seq_id == 'M95099'



import pytest
import random
import functools
from conftest import skip_no_connection

from mgkit.filter.common import FilterFails
from mgkit.filter.taxon import filter_taxon_by_id_list
from mgkit.taxon import is_ancestor


pytestmark = skip_no_connection

def test_taxon_id_list1(ncbi_taxonomy):
    # at least prevotella as genus should be amog the others
    taxon_id = ncbi_taxonomy.find_by_name('prevotella')[0]
    filter_list = [
        taxon.taxon_id
        for taxon in ncbi_taxonomy
        if 'prevotella' in taxon.s_name.lower()
    ]
    assert filter_taxon_by_id_list(taxon_id, filter_list)


def test_taxon_id_list1_anc(ncbi_taxonomy):
    # at least prevotella as genus should be among the others
    taxon_id = random.choice(
        [
            taxon.taxon_id
            for taxon in ncbi_taxonomy
            if taxon.s_name.lower().startswith('prevotella ')
        ]
    )
    filter_list = ncbi_taxonomy.find_by_name('prevotella')
    assert filter_taxon_by_id_list(
        taxon_id,
        filter_list,
        func=functools.partial(is_ancestor, ncbi_taxonomy)
    )


def test_taxon_id_list2(ncbi_taxonomy):
    # no prevotella species will be found
    filter_list = ncbi_taxonomy.find_by_name('prevotella')
    taxon_id = random.choice(
        [
            taxon.taxon_id
            for taxon in ncbi_taxonomy
            if ('prevotella' in taxon.s_name) and (taxon.rank != 'genus')
        ]
    )
    assert not filter_taxon_by_id_list(taxon_id, filter_list)


def test_taxon_id_list2_anc(ncbi_taxonomy):
    # at least prevotella as genus should be among the others
    taxon_id = 839
    filter_list = ncbi_taxonomy.find_by_name('prevotella')
    assert not filter_taxon_by_id_list(
        taxon_id,
        filter_list,
        exclude=True,
        func=functools.partial(is_ancestor, ncbi_taxonomy)
    )


def test_taxon_id_list_exception():
    # at least prevotella as genus should be among the others
    taxon_id = 839
    filter_list = None
    with pytest.raises(FilterFails):
        filter_taxon_by_id_list(
            taxon_id,
            filter_list,
            exclude=True,
            func=None
        )



import pytest


def test_filter_gff_version(script_runner):
    assert script_runner.run('filter-gff', '--version').success


def test_add_gff_info_version(script_runner):
    assert script_runner.run('add-gff-info', '--version').success


def test_get_gff_info_version(script_runner):
    assert script_runner.run('get-gff-info', '--version').success


def test_hmmer2gff_version(script_runner):
    assert script_runner.run('hmmer2gff', '--version').success


def test_blast2gff_version(script_runner):
    assert script_runner.run('blast2gff', '--version').success


def test_snp_parser_version(script_runner):
    assert script_runner.run('snp_parser', '--version').success


def test_fastq_utils_version(script_runner):
    assert script_runner.run('fastq-utils', '--version').success


def test_taxon_utils_version(script_runner):
    assert script_runner.run('taxon-utils', '--version').success


def test_json2gff_version(script_runner):
    assert script_runner.run('json2gff', '--version').success


def test_fasta_utils_version(script_runner):
    assert script_runner.run('fasta-utils', '--version').success


def test_sampling_utils_version(script_runner):
    assert script_runner.run('sampling-utils', '--version').success


def test_blast2gff_blastdb1(script_runner, shared_datadir):
    tab_file = str(shared_datadir / 'blast-outfmt6.tsv')
    assert script_runner.run('blast2gff', 'blastdb', '-n', tab_file, '-').success



# coding=utf8
"""
The module contains translation tables

Not all genetic codes are included, taken from:
`<http://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi?mode=t#SG2>`_
"""

UNIVERSAL = {
    'TTT': 'F', 'TCT': 'S', 'TAT': 'Y', 'TGT': 'C',
    'TTC': 'F', 'TCC': 'S', 'TAC': 'Y', 'TGC': 'C',
    'TTA': 'L', 'TCA': 'S', 'TAA': '*', 'TGA': '*',
    'TTG': 'L', 'TCG': 'S', 'TAG': '*', 'TGG': 'W',
    'CTT': 'L', 'CCT': 'P', 'CAT': 'H', 'CGT': 'R',
    'CTC': 'L', 'CCC': 'P', 'CAC': 'H', 'CGC': 'R',
    'CTA': 'L', 'CCA': 'P', 'CAA': 'Q', 'CGA': 'R',
    'CTG': 'L', 'CCG': 'P', 'CAG': 'Q', 'CGG': 'R',
    'ATT': 'I', 'ACT': 'T', 'AAT': 'N', 'AGT': 'S',
    'ATC': 'I', 'ACC': 'T', 'AAC': 'N', 'AGC': 'S',
    'ATA': 'I', 'ACA': 'T', 'AAA': 'K', 'AGA': 'R',
    'ATG': 'M', 'ACG': 'T', 'AAG': 'K', 'AGG': 'R',
    'GTT': 'V', 'GCT': 'A', 'GAT': 'D', 'GGT': 'G',
    'GTC': 'V', 'GCC': 'A', 'GAC': 'D', 'GGC': 'G',
    'GTA': 'V', 'GCA': 'A', 'GAA': 'E', 'GGA': 'G',
    'GTG': 'V', 'GCG': 'A', 'GAG': 'E', 'GGG': 'G'
}
"""
Universal genetic code
transl_table=1
"""

VT_MIT = UNIVERSAL.copy()
VT_MIT.update(
    {'AGA': '*', 'AGG': '*', 'ATA': 'M', 'TGA': 'W'}
)
"""
Vertebrate Mithocondrion
transl_table=2
"""

YST_MIT = UNIVERSAL.copy()
YST_MIT.update(
    {'ATA': 'M', 'CTT': 'T', 'CTC': 'T', 'CTA': 'T', 'CTG': 'T',
     'TGA': 'W'}
)
"""
Yeast Mithocondrion
transl_table=3
"""
# codons not in Yeast Mithocondrion:
del YST_MIT['CGA']
del YST_MIT['CGC']

PRT_MIT = UNIVERSAL.copy()
"""
The Mold, Protozoan, and Coelenterate Mitochondrial Code and the
Mycoplasma/Spiroplasma Code
transl_table=4
"""
PRT_MIT.update({'TGA': 'W'})

INV_MIT = UNIVERSAL.copy()
"""
The Invertebrate Mitochondrial Code
transl_table=5
"""
INV_MIT.update(
    {'AGA': 'S', 'AGG': 'S', 'ATA': 'M', 'TGA': 'W'}
)

DRS_MIT = INV_MIT.copy()
"Drosophila Mithocondrion genome lacks a codon, compare to vertebrates"
del DRS_MIT['AGG']

BAC_PLT = {
    'TTT': 'F', 'TCT': 'S', 'TAT': 'Y', 'TGT': 'C',
    'TTC': 'F', 'TCC': 'S', 'TAC': 'Y', 'TGC': 'C',
    'TTA': 'L', 'TCA': 'S', 'TAA': '*', 'TGA': '*',
    'TTG': 'L', 'TCG': 'S', 'TAG': '*', 'TGG': 'W',
    'CTT': 'L', 'CCT': 'P', 'CAT': 'H', 'CGT': 'R',
    'CTC': 'L', 'CCC': 'P', 'CAC': 'H', 'CGC': 'R',
    'CTA': 'L', 'CCA': 'P', 'CAA': 'Q', 'CGA': 'R',
    'CTG': 'L', 'CCG': 'P', 'CAG': 'Q', 'CGG': 'R',
    'ATT': 'I', 'ACT': 'T', 'AAT': 'N', 'AGT': 'S',
    'ATC': 'I', 'ACC': 'T', 'AAC': 'N', 'AGC': 'S',
    'ATA': 'I', 'ACA': 'T', 'AAA': 'K', 'AGA': 'R',
    'ATG': 'M', 'ACG': 'T', 'AAG': 'K', 'AGG': 'R',
    'GTT': 'V', 'GCT': 'A', 'GAT': 'D', 'GGT': 'G',
    'GTC': 'V', 'GCC': 'A', 'GAC': 'D', 'GGC': 'G',
    'GTA': 'V', 'GCA': 'A', 'GAA': 'E', 'GGA': 'G',
    'GTG': 'V', 'GCG': 'A', 'GAG': 'E', 'GGG': 'G'
}
"""
The Bacterial and Plant Plastid Code
transl_table=11
"""

YST_ALT = UNIVERSAL.copy()
"""
The Alternative Yeast Nuclear Code
transl_table=12
"""
YST_ALT.update({'CTG': 'S'})



"""
.. versionadded:: 0.3.3

GLM models with metagenomes and metatranscriptomes. Experimental
"""

from __future__ import division
from builtins import range
import statsmodels.api as sm
from scipy import interpolate, optimize
import pandas as pd


def lowess_ci_bootstrap(endog, exog, num=100, frac=.2, it=3, alpha=.05,
                        delta=0., min_value=10**-3, kind='slinear'):
    """
    .. versionchanged: 0.4.0
        the interpolation of the confidence intervals uses the passed `kind`

    Bootstraps a lowess for the dependent (`endog`) and indipendent (`exog`)
    arguments.

    Arguments:
        endog (array): indipendent variable (Y)
        exog (array): indipendent variable (X)
        num (int): number of iterations for the bootstrap
        frac (float): fraction of the array to use when fitting
        it (int): number of iterations used to fit the lowess
        alpha (float): confidence intervals for the bootstrap
        delta (float): passed to :func:`statsmodels.api.nonparametric.lowess`
        min_value (float): minimum value for the function to avoid out of
            bounds
        kind (str): type of interpolation passed to
            :func:`scipy.interpolate.interp1d`

    Returns:
        tuple: the first element is the function describing the lowest
        confidence interval, the second element is for the highest confidence
        interval and the last one for the mean

    .. note::

        Performance increase with the value of *delta*.
    """
    data = pd.DataFrame(
        {
            'endog': endog,
            'exog': exog
        }
    ).sort_values(by=['exog', 'endog'], ascending=True)

    boots = []
    for idx in range(num):
        sample = data.sample(n=data.index.size, replace=True)
        lw = sm.nonparametric.lowess(
            sample.endog,
            sample.exog,
            is_sorted=False,
            frac=frac,
            it=it,
            return_sorted=True,
            delta=delta
        )
        boots.append(pd.DataFrame({
            'endog': lw[:, 1],
            'exog': lw[:, 0]
        }))

    alpha = alpha / 2
    boots = pd.concat(boots)
    q1 = boots.groupby('exog').quantile(
        alpha, interpolation='nearest'
    ).sort_index()
    q1 = q1.endog[q1.endog > 0].fillna(min_value)
    q1 = interpolate.interp1d(q1.index, q1, kind=kind,
                              fill_value='extrapolate')

    q2 = boots.groupby('exog').quantile(
        1 - alpha, interpolation='nearest'
    ).sort_index()
    q2 = q2.endog[q2.endog > 0].fillna(min_value)
    q2 = interpolate.interp1d(q2.index, q2, kind=kind,
                              fill_value='extrapolate')

    m = fit_lowess_interpolate(data.endog, data.exog, frac=frac, it=it,
                               kind=kind)

    return q1, q2, m


def fit_lowess_interpolate(endog, exog, frac=.2, it=3, kind='slinear'):
    """
    Fits a lowess for the passed `endog` (Y) and `exog` (X) and returns an
    interpolated function that describes it. The first 4 arguments are passed
    to :func:`statsmodels.api.sm.nonparametric.lowess`, while the last one is
    passed to :func:`scipy.interpolate.interp1d`

    Arguments:
        endog (array): array of the dependent variable (Y)
        exog (array): array of the indipendent variable (X)
        frac (float): fraction of the number of elements to use when fitting
            (0.0-1.0)
        it (int): number of iterations to fit the lowess
        kind (str): type of interpolation to use

    Returns:
        func: interpolated function representing the lowess fitted from the
        data passed
    """

    data = pd.DataFrame(
        {
            'endog': endog,
            'exog': exog
        }
    ).sort_values(by=['exog', 'endog'], ascending=True)

    lw = sm.nonparametric.lowess(
        data.endog,
        data.exog,
        frac=frac,
        it=it,
    )

    return interpolate.interp1d(
        lw[:, 0],
        lw[:, 1],
        kind=kind,
        fill_value='extrapolate'
    )


def variance_to_alpha(mu, func, min_alpha=10**-3):
    """
    Based on the variance defined in the Negative Binomial in statsmodels

    var = mu + alpha * (mu ** 2)

    Arguments:
        mu (float): mean to calculate the alphas for
        func (func): function that returns the variace of the mean
        min_alpha (float): value of alpha if the `func` goes out of bounds

    Returns:
        float: value of alpha for the passed mean
    """

    alpha = (func(mu) - mu) / (mu ** 2)

    return min_alpha if alpha <= 0 else alpha


def optimise_alpha_scipy_function(args, formula, data, criterion='aic'):
    """
    .. versionadded:: 0.4.0


    """
    alpha, = args
    model = sm.formula.glm(
        formula,
        data=data,
        family=sm.families.NegativeBinomial(alpha=alpha)
    ).fit()
    return getattr(model, criterion)


def optimise_alpha_scipy(formula, data, mean_func, q1_func, q2_func):
    """
    .. versionadded:: 0.4.0

    Used to find an optimal *alpha* parameter for the Negative Binomial
    distribution used in `statsmodels`, using the lowess functions from
    :func:`lowess_ci_bootstrap`.

    Arguments:
        formula (str): the formula used for the regression
        data (DataFrame): DataFrame for regression
        mean_func (func): function for the mean :func:`lowess_ci_bootstrap`
        q1_func (func): function for the q1 :func:`lowess_ci_bootstrap`
        q2_func (func): function for the q2 :func:`lowess_ci_bootstrap`

    Returns:
        float: *alpha* value for the Negative Binomial
    """
    data_mean = data.mean()[0]
    return optimize.minimize(
        optimise_alpha_scipy_function,
        [variance_to_alpha(data_mean, mean_func)],
        args=(formula, data),
        bounds=[(variance_to_alpha(data_mean, q1_func), variance_to_alpha(data_mean, q2_func))]
    ).x[0]



"""
The script contains commands used to access functionality related to
taxonomy, without the need to write ad-hoc code for functionality that
can be part of a workflow. One example is access to the the last common
ancestor function contained in the :mod:`mgkit.taxon`.

Last Common Ancestor (lca and lca_line)
***************************************

These commands expose the functionality of
:func:`last_common_ancestor_multiple`, making it accessible via the command
line. They differ in the input file format and the choice of output files.

the *lca* command can be used to define the last common ancestor of contigs
from the annotation in a GFF file. The command uses the *taxon_ids* from all
annotations belonging to a contig/sequence, if they have a **bitscore** higher
or equal to the one passed (50 by default). The default output of the command
is a tab separated file where the first column is the contig/sequence name,
the taxon_id of the last common ancestor, its scientific/common name and its
lineage.

For example::

    contig_21   172788  uncultured phototrophic eukaryote   cellular organisms,environmental samples

If the *-r* is used, by passing the fasta file containing the nucleotide
sequences the output file is a GFF where for each an annotation for the full
contig length contains the same information of the tab separated file format.

The **lca_line** command accept as input a file where each line consist of a
list of taxon_ids. The separator for the list can be changed and it defaults to
TAB. The last common ancestor for all taxa on a line is searched. The ouput of
this command is the same as the tab separated file of the **lca** command, with
the difference that instead of the first column, which in this command becames
a list of all *taxon_ids* that were used to find the last common ancestor for
that line. The list of *taxon_ids* is separated by semicolon ";".

.. note::

    Both also accept the *-n* option, to report the config/line and the
    taxon_ids that had no common ancestors. These are treated as errors and do
    not appear in the output file.

Krona Output
############

.. versionadded:: 0.3.0

The *lca* command supports the writing of a file compatible with Krona. The
output file can be used with the *ktImportText/ImportText.pl* script included
with `KronaTools <https://github.com/marbl/Krona/wiki>`_. Specifically, the
output from *taxon_utils* will be a file with all the lineages found (tab
separated), that can be used with::

    $ ktImportText -q taxon_utils_ouput

Note the use of *-q* to make the script count the lineages. Sequences with no
LCA found will be marked as *No LCA* in the graph, the *-n* is not required.

.. note::

    Please note that the output won't include any sequence that didn't have a
    hit with the software used. If that's important, the **-kt** option can be
    used to add a number of *Unknown* lines at the end, to read the total
    supplied.

Filter by Taxon
***************

The **filter** command of this script allows to filter a GFF file using the
*taxon_id* attribute to include only some annotations, or exclude some. The
filter is based on the `mgkit.taxon.is_ancestor` function, and the
`mgkit.filter.taxon.filter_taxon_by_id_list`. It can also filter a table (tab
separated values) when the first element is an ID and the second is a taxon_id.
An example of a table of this sort is the output of the `download-ncbi-taxa.sh`
and `download-uniprot-taxa.sh`, where each accession of a database is associated
to a taxon_id.

Multiple taxon_id can be passed, either for inclusion or exclusion. If both
exclusion and inclusion is used, the first check is on the inclusion and then on
the exclusion. In alternative to passing taxon_id, taxon_names can be passed,
with values such as 'cellular organisms' that needs to be quoted. Example::

    $ taxon-utils filter -i 2 -in archaea -en prevotella -t taxonomy.pickle in.gff out.gff

Which will keep only line that are from Bacteria (taxon_id=2) and exclude those
from the genus *Prevotella*. It will be also include Archaea.

Multiple inclusion and exclusion flags can be put::

    $ taxon-utils filter -i 2 -i 2172 -t taxonomy in.gff out.gff

In particular, the inclusion flag is tested first and then the exclusion is
tested. So a line like this one:

.. code-block:: bash

    printf "TEST\\t838\\nTEST\\t1485" | taxon-utils filter -p -t taxonomy.pickle -i 2 -i 1485 -e 838

Will produce **TEST 1485**, because both Prevotella (838) and Clostridium (1485)
are Bacteria (2) OR Prevotella, but Prevotella must be excluded according to
the exclusion option. This line also illustrate that a tab-separated file, where
the second column contains taxon IDs, can be filtered. In particular it can be
applied to files produced by `download-ncbi-taxa.sh` or
`download-uniprot-taxa.sh` (see :ref:`download-taxonomy`).

.. warning::

    Annotations with no taxon_id are not included in the output of both filters

Convert Taxa Tables to HDF5
***************************

This command is used to convert the taxa tables download from Uniprot and NCBI,
using the scripts mentioned in :ref:`download-data`,
`download-uniprot-taxa.sh` and `download-ncbi-taxa` into a HDF5 file that can
be used with the *addtaxa* command in :ref:`add-gff-info`.

The advantage is a faster lookup of the IDs. The other is a smaller memory
footprint when a great number of annotations are kept in memory.

Extract taxonomy
****************

This command allows to print the taxonomy in a file created with MGKit. The
default behaviour is to print the scientific name, taxon_id, rank and lineage
(only ranked taxa) in a tab separated file (or standard output).

Besides the option to change the column separator, it is possible to only
print specific taxa (case insensitive search, but need the correct name) and
print headers for the output table.

Changes
*******

.. versionchanged:: 0.5.0
    added *get* command to taxon-utils to print the taxonomy or search in it

.. versionchanged:: 0.3.4
    changed interface and behaviour for *filter*, also now can filter tables;
    *lca* has changed the interface and allows the output of a 2 column table

.. versionchanged:: 0.3.1
    added *to_hdf* command

.. versionchanged:: 0.3.1
    added *-j* option to *lca*, which outputs a JSON file with the LCA results

.. versionchanged:: 0.3.0
    added *-k* and *-kt* options for Krona output, lineage now includes the LCA
    also added *-a* option to select between lineages with only ranked taxa.
    Now it defaults to all components.

.. versionchanged:: 0.2.6
    added *feat-type* option to *lca* command, added phylum output to nolca

.. versionadded:: 0.2.5

"""
from __future__ import division
from builtins import range
from future.utils import viewvalues
import logging
import functools
import json
import click
import mgkit
import pandas as pd
from tqdm import tqdm
from . import utils
from mgkit.io import gff, fasta, blast
from mgkit import taxon
from mgkit.simple_cache import memoize
from mgkit.filter.taxon import filter_taxon_by_id_list


LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


def write_lca_gff(file_handle, seq_id, seq, taxon_id, taxon_name, lineage,
                  feat_type):
    annotation = gff.from_sequence(
        seq_id,
        seq,
        feat_type=feat_type,
        taxon_id=taxon_id,
        taxon_name=taxon_name,
        lineage=lineage
    )
    annotation.to_file(file_handle)


def write_lca_tab(file_handle, seq_id, taxon_id, taxon_name, rank, lineage):
    file_handle.write(
        "{}\t{}\t{}\t{}\t{}\n".format(
            seq_id,
            taxon_id,
            taxon_name,
            rank,
            lineage
        ).encode('ascii')
    )


def write_lca_tab_simple(file_handle, seq_id, taxon_id):
    file_handle.write(
        "{}\t{}\n".format(seq_id, taxon_id).encode('ascii')
    )


def get_taxon_info(taxonomy, taxon_id, only_ranked):
    if taxonomy[taxon_id].s_name:
        taxon_name = taxonomy[taxon_id].s_name
    else:
        taxon_name = taxonomy[taxon_id].c_name
    lineage = ','.join(
        tx
        for tx in taxon.get_lineage(taxonomy, taxon_id, names=True, only_ranked=only_ranked, with_last=True)
        if tx
    )
    return taxon_name, lineage


def write_no_lca(file_handle, seq_id, taxon_ids, extra=None):
    file_handle.write(
        "{}\t{}\t{}\n".format(
            seq_id,
            ','.join(str(taxon_id) for taxon_id in set(taxon_ids)),
            '' if extra is None else ','.join(extra)
        ).encode('ascii')
    )


def write_krona(file_handle, taxonomy, taxon_id, only_ranked):
    if taxon_id is None:
        lineage = ['No LCA']
    else:
        lineage = taxon.get_lineage(
            taxonomy,
            taxon_id,
            names=True,
            only_ranked=only_ranked,
            with_last=True
        )
    file_handle.write(
        '{}\n'.format('\t'.join(lineage)).encode('ascii')
    )


def write_json(lca_dict, seq_id, taxonomy, taxon_id, only_ranked):
    if taxon_id is None:
        lineage = ""
    else:
        lineage = taxon.get_lineage(
            taxonomy,
            taxon_id,
            names=True,
            only_ranked=only_ranked,
            with_last=True
        )

    lca_dict[seq_id] = dict(
        taxon_id=taxon_id,
        lineage=';'.join(lineage),
    )


@main.command('lca', help="""Finds the last common ancestor for each sequence
              in a GFF file""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--taxonomy', required=True, help='Taxonomy file')
@click.option('-n', '--no-lca', type=click.File('wb'), default=None,
              help='File to which write records with no LCA')
@click.option('-a', '--only-ranked', is_flag=True, default=False,
              help='''If set, only taxa that have a rank will be used in the lineageself. This is not advised for lineages such as Viruses, where the top levels have no rank''')
@click.option('-b', '--bitscore', default=0, type=click.FLOAT, show_default=True,
              help='Minimum bitscore accepted')
@click.option('-m', '--rename', is_flag=True, default=False,
              help='Emulates BLAST behaviour for headers (keep left of first space)')
@click.option('-s', '--sorted', is_flag=True, default=False,
              help='''If the GFF file is sorted (all of a sequence annotations are contiguos) can use less memory, `sort -s -k 1,1` can be used''')
@click.option('-ft', '--feat-type', default='LCA', show_default=True,
              help='Feature type used if the output is a GFF (default is *LCA*)')
@click.option('-r', '--reference', default=None, type=click.File('rb'),
              help='Reference file for the GFF, if supplied a GFF file is the output')
@click.option('-p', '--simple-table', is_flag=True, default=False,
              help='Uses a 2 column table format (seq_id taxon_id) TAB separated')
@click.option('-kt', '--krona-total', type=click.INT, default=None,
              help='''Total number of raw sequences (used to output correct percentages in Krona''')
@click.option('-f', '--out-format', default='tab', show_default=True,
              type=click.Choice(['krona', 'json', 'tab', 'gff']),
              help='Format of output file')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('gff-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def lca_contig_command(verbose, taxonomy, no_lca, only_ranked, bitscore,
                       rename, sorted, feat_type, reference, simple_table,
                       krona_total, out_format, progress, gff_file,
                       output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    if (out_format == 'gff') and (reference is None):
        utils.exit_script(
            'The output format *gff* requires a FASTA file, passed with -r', 3)

    if no_lca is not None:
        LOG.info(
            "Writing contigs/taxon_ids of contigs with no LCA to (%s)",
            getattr(no_lca, 'name', repr(no_lca))
        )

    taxonomy = taxon.Taxonomy(taxonomy)

    if reference is not None:
        seqs = dict(
            fasta.load_fasta_rename(reference) if rename else fasta.load_fasta(reference)
        )
    else:
        seqs = None

    # basic filter for the presence of a taxon_id and bitscore
    annotations_iter = (
        annotation
        for annotation in gff.parse_gff(gff_file)
        # only use annotations whose bitscore pass the filter
        # and have a taxon_id
        if (
            (annotation.bitscore >= bitscore) or
            (annotation.bitscore is None)
            ) and (annotation.taxon_id is not None) and
            # redundant probably, but used in cases when a taxon_id was deleted
            # from the taxonomy
            (annotation.taxon_id in taxonomy)
    )

    if sorted:
        LOG.info("Input GFF is assumed sorted")
        annotations = gff.group_annotations_sorted(
            annotations_iter,
            lambda annotation: annotation.seq_id
        )
    else:
        # groups the annotations by sequence, in case they're not sorted
        annotations = viewvalues(gff.group_annotations(
            annotations_iter,
            lambda annotation: annotation.seq_id
        ))

    count = 0
    lca_dict = {}
    if progress:
        annotations = tqdm(annotations)
    for seq_ann in annotations:
        count += 1
        seq_id = seq_ann[0].seq_id
        try:
            taxon_id = taxon.last_common_ancestor_multiple(
                taxonomy,
                (annotation.taxon_id for annotation in seq_ann)
            )
        except taxon.NoLcaFound as error:
            LOG.warning("No LCA found for %s (%s)", seq_id, error)
            if no_lca is not None:
                write_no_lca(
                    no_lca,
                    seq_id,
                    (annotation.taxon_id for annotation in seq_ann),
                    extra=set(
                        taxonomy.get_ranked_taxon(
                            annotation.taxon_id,
                            'phylum'
                        ).s_name
                        for annotation in seq_ann
                    )
                )
            if out_format == 'krona':
                write_krona(output_file, taxonomy, None, False)
            continue

        taxon_name, lineage = get_taxon_info(
            taxonomy,
            taxon_id,
            only_ranked
        )
        if out_format == 'gff':
            write_lca_gff(
                output_file,
                seq_id,
                seqs[seq_id],
                taxon_id,
                taxon_name,
                lineage,
                feat_type
            )
        elif out_format == 'krona':
            write_krona(
                output_file,
                taxonomy,
                taxon_id,
                only_ranked
            )
        elif out_format == 'json':
            write_json(
                lca_dict,
                seq_id,
                taxonomy,
                taxon_id,
                only_ranked
            )
        elif out_format == 'tab':
            if simple_table:
                write_lca_tab_simple(output_file, seq_id, taxon_id)
            else:
                write_lca_tab(
                    output_file,
                    seq_id,
                    taxon_id,
                    taxon_name,
                    taxonomy[taxon_id].rank,
                    lineage
                )
    if (out_format == 'krona') and (krona_total is not None):
        for index in range(count, krona_total):
            output_file.write('Unknown\n')

    if out_format == 'json':
        output_file.write(json.dumps(lca_dict, indent=4).encode('ascii'))


@main.command('lca_line', help="""Finds the last common ancestor for all IDs in
              a text file line""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--taxonomy', required=True, help='Taxonomy file')
@click.option('-n', '--no-lca', type=click.File('wb'), default=None,
              help='File to which write records with no LCA')
@click.option('-a', '--only-ranked', is_flag=True, default=False,
              help='''If set, only taxa that have a rank will be used in the lineageself. This is not advised for lineages such as Viruses, where the top levels have no rank''')
@click.option('-s', '--separator', default='\t',
              help='separator for taxon_ids (defaults to TAB)')
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def lca_line_command(verbose, taxonomy, no_lca, only_ranked, separator,
                     input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    if no_lca is not None:
        LOG.info(
            "Writing taxon_ids of lines with no LCA to (%s)",
            getattr(no_lca, 'name', repr(no_lca))
        )

    taxonomy = taxon.Taxonomy(taxonomy)

    for line in input_file:
        line = line.decode('ascii')
        taxon_ids = [
            int(taxon_id)
            for taxon_id in line.strip().split(separator)
        ]
        try:
            taxon_id = taxon.last_common_ancestor_multiple(
                taxonomy,
                taxon_ids
            )
        except taxon.NoLcaFound as error:
            LOG.error("No LCA found for %s (%s)", taxon_ids, error)
            if no_lca is not None:
                write_no_lca(
                    no_lca,
                    '',
                    taxon_ids
                )
            continue
        taxon_name, lineage = get_taxon_info(taxonomy, taxon_id, only_ranked)
        write_lca_tab(
            output_file,
            ';'.join(str(x) for x in taxon_ids),
            taxon_id,
            taxon_name,
            taxonomy[taxon_id].rank,
            lineage
        )


def validate_taxon_ids(taxon_ids, taxonomy):
    taxon_ids = set(taxon_ids)
    broken = False
    for taxon_id in taxon_ids:
        if taxon_id not in taxonomy:
            LOG.critical(
                "taxon_id %d was not found in the taxonomy",
                taxon_id
            )
            broken = True
    if broken:
        utils.exit_script(
            "Some of taxon_ids were not found", 1
        )
    return taxon_ids


def validate_taxon_names(taxon_names, taxonomy):

    broken = False
    taxon_ids = set()
    for taxon_name in taxon_names:
        try:
            taxon_id = taxonomy.find_by_name(taxon_name)
        except KeyError:
            LOG.critical("'%s' was not found in the taxonomy", taxon_name)
            broken = True
            continue
        if len(taxon_id) > 1:
            LOG.critical(
                "'%s' was found multiple times in the taxonomy: (%s)",
                taxon_name,
                ', '.join(str(tid) for tid in taxon_id)
            )
            broken = True
            continue
        # takes the first (and only) taxon_id from the list returned
        taxon_ids.add(taxon_id[0])

    if broken:
        utils.exit_script(
            "Some of taxon_names were not found", 2
        )
    return taxon_ids


@main.command('filter', help="Filter a GFF file or a table based on taxonomy")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-p', '--table', is_flag=True, default=False)
@click.option('-t', '--taxonomy', required=True, help='Taxonomy file')
@click.option('-i', '--include-taxon-id', multiple=True, default=None,
              type=click.INT, help='Include only taxon_ids')
@click.option('-in', '--include-taxon-name', multiple=True, default=None,
              help='Include only taxon_names')
@click.option('-e', '--exclude-taxon-id', multiple=True, default=None,
              type=click.INT, help='Exclude taxon_ids')
@click.option('-en', '--exclude-taxon-name', multiple=True, default=None,
              help='Exclude taxon_names')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def filter_taxa_command(verbose, table, taxonomy, include_taxon_id,
                        include_taxon_name, exclude_taxon_id,
                        exclude_taxon_name, progress, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    taxonomy = taxon.Taxonomy(taxonomy)

    exclude_ids = validate_taxon_ids(exclude_taxon_id, taxonomy) | \
        validate_taxon_names(exclude_taxon_name, taxonomy)

    include_ids = validate_taxon_ids(include_taxon_id, taxonomy) | \
        validate_taxon_names(include_taxon_name, taxonomy)

    if exclude_ids:
        LOG.info("Excluding Taxa: %s", exclude_ids)
        exclude_func = functools.partial(
            filter_taxon_by_id_list,
            filter_list=exclude_ids,
            exclude=True,
            func=functools.partial(
                taxon.is_ancestor,
                taxonomy
            )
        )
        exclude_func = memoize(exclude_func)
    else:
        exclude_func = None
    if include_ids:
        LOG.info("Only include Taxa: %s", include_ids)
        include_func = functools.partial(
            filter_taxon_by_id_list,
            filter_list=include_ids,
            exclude=False,
            func=functools.partial(
                taxon.is_ancestor,
                taxonomy
            )
        )
        include_func = memoize(include_func)
    else:
        include_func = None

    if table:
        iterator = blast.parse_accession_taxa_table(input_file, key=0, value=1,
                                                    num_lines=None)
        if progress:
            iterator = tqdm(iterator)
        for acc_id, taxon_id in iterator:
            if include_func is not None:
                if not include_func(taxon_id):
                    continue
            if exclude_func is not None:
                if not exclude_func(taxon_id):
                    continue
            output_file.write("{}\t{}\n".format(acc_id, taxon_id).encode('ascii'))
    else:
        iterator = gff.parse_gff(input_file)
        if progress:
            iterator = tqdm(iterator)
        for annotation in iterator:
            if annotation.taxon_id is None:
                continue
            if include_func is not None:
                if not include_func(annotation.taxon_id):
                    continue
            if exclude_func is not None:
                if not exclude_func(annotation.taxon_id):
                    continue
            annotation.to_file(output_file)


@main.command('to_hdf', help="""Convert a taxa table to HDF5, with the input as
              tabular format, defaults to stdin. Output file, defaults to
              (taxa-table.hf5)""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-n', '--table-name', default='taxa', show_default=True,
              help='Name of the table/storage to use')
@click.option('-w', '--overwrite', default=False, is_flag=True,
              help='Overwrite the file, instead of appending to it')
@click.option('-s', '--index-size', default=12, type=click.INT,
              show_default=True,
              help='Maximum number of characters for the gene_id')
@click.option('-c', '--chunk-size', default=5000000, type=click.INT,
              show_default=True,
              help='Chunk size to use when reading the input file')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input_file', type=click.File('rb'), default='-')
@click.argument('output_file', default='taxa-table.hf5')
def taxa_table_command(verbose, table_name, overwrite, index_size, chunk_size,
                       progress, input_file, output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    hdf = pd.HDFStore(
        output_file,
        mode='w' if overwrite else 'a'
    )

    iterator = pd.read_table(
        input_file, header=None,
        squeeze=False,
        index_col=0,
        engine='c',
        chunksize=chunk_size,
        names=['taxon_id']
    )
    LOG.info('Reading Taxa Table from file (%s)', input_file.name)
    LOG.info(
        'Writing HDF5 file (%s) Table (%s) - Overwrite: %s',
        output_file,
        table_name,
        overwrite
    )
    if progress:
        iterator = tqdm(iterator)
    for chunk in iterator:
        hdf.append(
            table_name,
            chunk,
            min_itemsize=index_size,
            index=False,
            # data_columns=False
        )

    LOG.info('Creating Indices')
    hdf.create_table_index(table_name, optlevel=9, kind='full')
    hdf.close()

    LOG.info("It's reccomended to compress the file with `ptrepack`")
    LOG.info(
        "e.g. ptrepack --propindexes --complevel 9 --complib blosc %s:/ taxa-table-compressed.hf5:/",
        output_file
    )


def output_taxon_line(taxonomy, taxon_id, sep='\t'):
    taxon_name = taxonomy[taxon_id].s_name
    lineage = taxonomy.get_lineage_string(taxon_id)
    rank = taxonomy[taxon_id].rank

    return sep.join([taxon_name, str(taxon_id), rank, lineage]) + '\n'


@main.command('get', help="""Extract the taxonomy as a CSV file""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-d', '--header', is_flag=True,
              help='Include header in the output')
@click.option('-s', '--separator', default='\t', help='column separator')
@click.option('-o', '--only-names', multiple=True, type=click.STRING,
              help='Only get matched taxon names')
@click.argument('taxonomy_file', type=click.File('rb'))
@click.argument('output_file', type=click.File('w'), default='-')
def get_taxonomy(verbose, header, separator, only_names, taxonomy_file,
                 output_file):
    """
    .. versionadded:: 0.5.0
    """
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    taxonomy = taxon.Taxonomy(taxonomy_file)

    if header:
        output_file.write(separator.join(
            ['Taxon Name', 'taxon_id', 'Rank', 'Lineage']) + '\n')

    if only_names:
        taxon_ids = []
        for taxon_name in only_names:
            try:
                taxon_ids.extend(taxonomy.find_by_name(taxon_name))
            except KeyError:
                LOG.error("Taxon '%s' not found", taxon_name)
        for taxon_id in taxon_ids:
            line = output_taxon_line(taxonomy, taxon_id, sep=separator)
            output_file.write(line)
    else:
        for taxon_obj in taxonomy:
            line = output_taxon_line(taxonomy, taxon_obj.taxon_id,
                                     sep=separator)
            output_file.write(line)



"""
Some test functions to filter sequences
"""
from __future__ import division
from builtins import range

import numpy

# from numpy import sum
# import pandas
# from scipy.stats import itemfreq
# import numexpr as ne

# ne.set_num_threads(8)


# #direct use: fourth (1.6ms loop)
# #in trim_by_ee: fourth
# def expected_error_rate_1(qualities):
#     qualities = pandas.DataFrame(itemfreq(qualities))

#     return qualities.prod(axis=1).sum()


# direct use: first (0.033ms loop)
# in trim_by_ee: first (15.9s)
def expected_error_rate(qualities):
    """
    Calculate the expected error rate for an array of qualities (converted to
    probabilities).
    """
    unique, inv = numpy.unique(qualities, return_inverse=True)
    return sum(unique * numpy.bincount(inv))


# #direct use: third (0.463ms loop)
# #in trim_by_ee: third (65.7s)
# def expected_error_rate_3(qualities):
#     return sum(
#         sum(1 for value in qualities if value == unique) * unique
#         for unique in numpy.unique(qualities)
#     )


# #direct use: second (0.109ms loop)
# #in trim_by_ee: second (27.7s)
# def expected_error_rate_4(qualities):
#     return sum(
#         len(qualities[qualities == unique]) * unique
#         for unique in numpy.unique(qualities)
#     )


def trim_by_ee(qualities, min_length=50, threshold=0.5, chars=True, base=33):
    """
    Trim a sequence based on the expected error rate.
    """
    if chars:
        qualities = numpy.fromiter(
            (ord(score) for score in qualities),
            dtype=int
        ) - base
    # qualities = ne.evaluate("10 ** ((-qualities) / 10)")
    qualities = 10 ** ((-qualities) / 10)
    # print qualities
    for idx in range(min_length, qualities.size):
        # print expected_error_rate(qualities[:idx])
        if expected_error_rate(qualities[:idx]) > threshold:
            return idx
    return qualities.size



"""
Script to convert HMMER results files (domain table) to a GFF file, the name of
the profiles are expected to be now in the form
*GENEID_TAXONID_TAXON-NAME(-nr)* by default, but any other profile name is
accepted.

The profiles tested are those made from Kegg Orthologs, from the
`download_profiles` script. If the `--no-custom-profiles` options is used,
the script can be used with any profile name. The profile name will be used
for `gene_id`, `taxon_id` and `taxon_name` in the GFF file.

It is possible to use seuqnces not translated using mgkit, no information on
the frame is assumed, so this script can be used against a protein DB. For
example Uniprot can be searched for profiles, in which case the **--no-frame**
options must be used.

.. note::

    for GENEID, old documentation points to KOID, it is the same

.. warning::

    The compatibility with old data has been **removed**, meaning that old
    experiments must use the scripts from those versions. It is possible to use
    multiple environments, with `virtualenv` for this purpose. An examples is
    given in :ref:`install-ref`.

Changes
*******

.. versionchanged:: 0.1.15
    adapted to new GFF module and specs

.. versionchanged:: 0.2.1
    added options to customise output and filters and old restrictions

.. versionchanged:: 0.3.1
    added *--no-frame* option for non mgkit-translated proteins, sequence
    headers are handled the same way as HMMER (truncated at the first space)

"""

import sys
import logging
import argparse
from mgkit import logger
from mgkit.io import gff
from mgkit.io import fasta
from . import utils

LOG = logging.getLogger(__name__)


def set_parser():
    """
    Setup command line options
    """
    parser = argparse.ArgumentParser(
        description='Convert HMMER data to GFF file',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    group = parser.add_argument_group('File options')
    group.add_argument(
        'aa_file',
        type=argparse.FileType('r'),
        help="Fasta file containing contigs translated to aa (used by HMMER)"
    )
    group.add_argument(
        'hmmer_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-'
    )
    group.add_argument(
        '-o',
        '--output-file',
        nargs='?',
        type=argparse.FileType('w'), default=sys.stdout
    )

    group = parser.add_argument_group('Filters')
    group.add_argument(
        '-t',
        '--discard',
        action='store',
        type=float,
        default=0.05,
        help='Evalue over which an hit will be discarded'
    )
    group.add_argument(
        '-d',
        '--disable-evalue',
        action='store_true',
        default=False,
        help='Disable Evalue filter'
    )

    group = parser.add_argument_group('GFF')
    group.add_argument(
        '-c',
        '--no-custom-profiles',
        action='store_false',
        default=True,
        help='Profiles names are not in the custom format'
    )
    group.add_argument(
        '-db',
        '--database',
        action='store',
        default='CUSTOM',
        help='Database from which the profiles are generated " +" (e.g. PFAM)'
    )
    group.add_argument(
        '-f',
        '--feature-type',
        action='store',
        default='gene',
        help='Type of feature (e.g. gene)'
    )
    group.add_argument(
        '-n',
        '--no-frame',
        action='store_true',
        default=False,
        help='Set if the sequences were not translated with translate_seq'
    )

    group = parser.add_argument_group('Misc')

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def get_aa_data(f_handle):
    """
    Load aminoacid seuqnces used by HMMER.
    """
    # LOG.info('Loading aa data from file %s', f_handle.name)

    aa_seqs = dict(
        (name.split(' ')[0], seq)
        for name, seq in fasta.load_fasta(f_handle)
    )

    return aa_seqs


def parse_domain_table_contigs(options):
    """
    Parse the HMMER result file
    """
    aa_seqs = get_aa_data(options.aa_file)

    LOG.info('Parsing HMMER data from file %s', options.hmmer_file.name)
    LOG.info('Writing GFF data to file %s', options.output_file.name)

    count_dsc = 0
    count_tot = 0
    count_skp = 0

    for idx, line in enumerate(options.hmmer_file):

        if line.startswith('#'):
            continue
        if idx % 10000 == 0:
            LOG.info("Line number: %d", idx)

        count_tot += 1

        try:
            annotation = gff.from_hmmer(
                line,
                aa_seqs,
                feat_type=options.feature_type,
                db=options.database,
                custom_profiles=options.no_custom_profiles,
                noframe=options.no_frame,
            )
        except ZeroDivisionError:
            LOG.error(
                "Skipping line %d because of an error in the calculations",
                idx + 1
            )
            count_skp += 1
            continue

        # if disable_evalue is True, skips filter
        if not options.disable_evalue:
            if annotation.score > options.discard:
                count_dsc += 1
                continue

        annotation.to_file(options.output_file)

    LOG.info(
        "Read %d lines, discarded %d, skipped %d",
        count_tot,
        count_dsc,
        count_skp
    )


def main():
    """
    Main loop
    """
    options = set_parser().parse_args()
    logger.config_log(options.verbose)

    parse_domain_table_contigs(
        options
    )


if __name__ == '__main__':
    main()



class memoize(dict):
    """
    a cache found on the `PythonDecoratorLibrary <https://wiki.python.org/moin/PythonDecoratorLibrary#Alternate_memoize_as_dict_subclass>`_

    Not sure about the license for it.
    """
    def __init__(self, func):
        self.func = func

    def __call__(self, *args):
        return self[args]

    def __missing__(self, key):
        result = self[key] = self.func(*key)
        return result



import pytest
from builtins import range
from mgkit.utils.dictionary import find_id_in_dict, combine_dict, \
    reverse_mapping, link_ids, combine_dict_one_value, merge_dictionaries, \
    split_dictionary_by_value, apply_func_to_values


@pytest.fixture
def valuedict():
    return {
        1: [1.0, 2.0],
        2: [2.0, 3.0],
        3: [4.0, 5.0, 6.0],
        4: [7.0]
    }


def test_combine_dict(valuedict):
    keydict = {
        'A': [1, 2, 3],
        'B': [3, 4]
    }
    result_dict = {
        'A': {1.0, 2.0, 3.0, 4.0, 5.0, 6.0},
        'B': {4.0, 5.0, 6.0, 7.0},
    }
    assert combine_dict(keydict, valuedict) == result_dict


def test_combine_dict_one_value(valuedict):
    keydict = {
        'A': 1,
        'B': 2
    }
    result_dict = {
        'A': [1.0, 2.0],
        'B': [2.0, 3.0],
    }
    assert combine_dict_one_value(keydict, valuedict) == result_dict


@pytest.fixture
def id_map():
    return {
        'k1': ['v2', 'v3'],
        'k2': ['v3'],
        'k3': ['v4', 'v1'],
        'k4': ['v2', 'v1'],
    }


def test_link_ids1(id_map):

    result_dict = {
        'k3': set(['k4']),
        'k2': set(['k1']),
        'k1': set(['k2', 'k4']),
        'k4': set(['k3', 'k1'])
    }

    assert link_ids(id_map) == result_dict


# Try to parametrise this another time
def test_link_ids2(id_map):

    result_dict = {
        'k3': set(['k4']),
        'k2': set([]),
        'k1': set(['k4']),
        'k4': set(['k3', 'k1'])
    }

    assert link_ids(id_map, black_list=['v3']) == result_dict


def test_reverse_mapping(id_map):
    result_dict = {
        'v1': set(['k3', 'k4']),
        'v2': set(['k1', 'k4']),
        'v3': set(['k2', 'k1']),
        'v4': set(['k3'])
    }
    assert reverse_mapping(id_map) == result_dict


def test_find_id_in_dict(id_map):
    assert sorted(find_id_in_dict('v1', id_map)) == ['k3', 'k4']


def test_merge_dictionaries():

    res = merge_dictionaries(
        [
            {
                1: range(5)
            },
            {
                1: range(8)
            },
            {
                1: 9
            }
        ]
    )
    assert res == {1: {0, 1, 2, 3, 4, 5, 6, 7, 9}}


@pytest.mark.parametrize(
    "threshold,key_filter,result_dict1,result_dict2",
    [
        (6, None, {1: [1.0, 2.0], 2: [2.0, 3.0]}, {3: [4.0, 5.0, 6.0], 4: [7.0]}),
        (5, None, {1: [1.0, 2.0]}, {2: [2.0, 3.0], 3: [4.0, 5.0, 6.0], 4: [7.0]}),
        (5, [1, 2, 4], {1: [1.0, 2.0]}, {2: [2.0, 3.0], 4: [7.0]}),
    ]
)
def test_split_dictionary_by_value(threshold, key_filter, result_dict1, result_dict2, valuedict):
    assert split_dictionary_by_value(valuedict, threshold, sum, key_filter) == (result_dict1, result_dict2)


def test_apply_func_to_values():
    d = {
        1: ('vg', 'vt')
    }

    assert apply_func_to_values(d, lambda x: x[0]) == {1: {'v'}}



"""
Blast routines and parsers

"""
import logging
from builtins import range, zip
from future.utils import viewitems
from . import gff
from . import open_file

NUM_LINES = 10 ** 6

LOG = logging.getLogger(__name__)


def add_blast_result_to_annotation(annotation, gi_taxa_dict, taxonomy,
                                   threshold=60):
    """
    .. deprecated:: 0.4.0

    Adds blast information to a GFF annotation.

    :param annotation: GFF annotation object
    :param dict gi_taxa_dict: dictionary returned by
        :func:`parse_gi_taxa_table`.
    :param taxonomy: Uniprot taxonomy, used to add the taxon name to the
        annotation
    """

    try:
        hit = gi_taxa_dict[annotation.attributes.ko_idx]
    except KeyError:
        return

    if hit.score <= threshold:
        return

    try:
        # skips if the ID is not in the Uniprot Taxonomy
        annotation.attributes.blast_taxon = taxonomy[hit.taxon_id].s_name
        annotation.attributes.blast_taxon_idx = hit.taxon_id
    except KeyError:
        return


def parse_blast_tab(file_handle, seq_id=0, ret_col=(0, 1, 2, 6, 7, 11),
                    key_func=None, value_funcs=None):
    """
    .. versionadded:: 0.1.12

    Parses blast output tab format, returning for each line a key (the query
    id) and the columns requested in a tuple.

    Arguments:
        file_handle (file): file name or file handle for the blast ouput
        seq_id (int): index for the column which has the query id
        ret_col (list, None): list of indexes for the columns to be returned or
            *None* if all columns must be returned
        key_func (None, func): function to transform the query id value in the
            key returned. If *None*, the query id is used
        value_funcs (None, list): list of functions to transform the value of
            all the requested columns. If *None* the values are not converted

    Yields:
        tuple: iterator of tuples with the first element being the query id
        after key_func is applied, if requested and the second element of
        the tuple is a tuple with the requested columns *ret_col*

    .. table:: BLAST+ used with `-outfmt 6`, default columns

        +--------------+--------------------------------+
        | column index |          description           |
        +==============+================================+
        |            0 | query name                     |
        +--------------+--------------------------------+
        |            1 | subject name                   |
        +--------------+--------------------------------+
        |            2 | percent identities             |
        +--------------+--------------------------------+
        |            3 | aligned length                 |
        +--------------+--------------------------------+
        |            4 | number of mismatched positions |
        +--------------+--------------------------------+
        |            5 | number of gap positions        |
        +--------------+--------------------------------+
        |            6 | query sequence start           |
        +--------------+--------------------------------+
        |            7 | query sequence end             |
        +--------------+--------------------------------+
        |            8 | subject sequence start         |
        +--------------+--------------------------------+
        |            9 | subject sequence end           |
        +--------------+--------------------------------+
        |           10 | e-value                        |
        +--------------+--------------------------------+
        |           11 | bit score                      |
        +--------------+--------------------------------+

    """

    if key_func is None:
        def key_func(x): return x

    if ret_col is None:
        ret_col = list(range(12))

    if value_funcs is None:
        value_funcs = [lambda x: x for y in range(len(ret_col))]

    if isinstance(file_handle, str):
        file_handle = open_file(file_handle, 'r')

    LOG.info(
        "Reading BLAST results from file (%s)",
        getattr(file_handle, 'name', repr(file_handle))
    )

    lineno = 0
    comments = 0

    for lineno, line in enumerate(file_handle):
        line = line.decode('ascii')
        if line.startswith('#'):
            comments += 1
            continue

        line = line.strip()

        if not line:
            continue

        line = line.split('\t')
        key = key_func(line[seq_id])
        values = tuple(
            func(line[index])
            for index, func in zip(ret_col, value_funcs)
        )

        yield key, values

    LOG.info('Read %d BLAST records', lineno + 1 - comments)


def parse_uniprot_blast(file_handle, bitscore=40, db='UNIPROT-SP', dbq=10,
                        name_func=None, feat_type='CDS', seq_lengths=None):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.1.13
        added *name_func* argument

    .. versionchanged:: 0.2.1
        added *feat_type*

    .. versionchanged:: 0.2.3
        added *seq_lengths* and added subject *start* and *end* and *e-value*

    Parses BLAST results in tabular format using :func:`parse_blast_tab`,
    applying a basic bitscore filter. Returns the annotations associated with
    each BLAST hit.

    Arguments:
        file_handle (str, file): file name or open file handle
        bitscore (int, float): the minimum bitscore for an annotation to be
            accepted
        db (str): database used
        dbq (int): an index indicating the quality of the sequence database
            used; this value is used in the filtering of annotations
        name_func (func): function to convert the name of the database
            sequences. Defaults to `lambda x: x.split('|')[1]`, which can be
            be used with fasta files provided by Uniprot
        feat_type (str): feature type in the GFF
        seq_lengths (dict): dictionary with the sequences lengths, used to
            deduct the frame of the '-' strand

    Yields:
        Annotation: instances of :class:`mgkit.io.gff.Annotation` instance of
        each BLAST hit.
    """

    if name_func is None:
        def name_func(x): return x.split('|')[1]

    ret_col = (0, 1, 2, 6, 7, 8, 9, 10, 11)

    # the second function extract the Uniprot ID from the sequence header
    value_funcs = (
        str,
        name_func,
        float,
        int,
        int,
        int,
        int,
        float,
        float
    )

    for seq_id, hit in parse_blast_tab(file_handle, ret_col=ret_col,
                                       value_funcs=value_funcs):
        if hit[-1] < bitscore:
            continue

        seq_len = None if seq_lengths is None else seq_lengths[seq_id]

        yield gff.from_nuc_blast(hit, db=db, dbq=dbq, feat_type=feat_type,
                                 seq_len=seq_len)


def parse_fragment_blast(file_handle, bitscore=40.0):
    """
    .. versionadded:: 0.1.13

    Parse the output of a BLAST output where the sequences are the single
    annotations, so the sequence names are the *uid* of the annotations.

    The only returned values are the best hits, maxed by bitscore and identity.

    Arguments:
        file_handle (str, file): file name or open file handle
        bitscore (float): minimum bitscore for accepting a hit

    Yields:
        tuple: a tuple whose first element is the *uid* (the sequence name) and
        the second is the a list of tuples whose first element is the GID (NCBI
        identifier), the second one is the identity and the third is the
        bitscore of the hit.

    """

    value_funcs = (lambda x: x.split('|')[1], float, float)

    iterator = parse_blast_tab(
        file_handle,
        ret_col=(1, 2, 11),
        value_funcs=value_funcs
    )

    uidmap = {}

    for uid, hit in iterator:

        if hit[-1] < bitscore:
            continue

        try:
            uidmap[uid].append(hit)
        except KeyError:
            uidmap[uid] = [hit]

    for uid, hits in viewitems(uidmap):
        # returns the hit with the max bitscore and max identity
        yield uid, hits


def parse_accession_taxa_table(file_handle, acc_ids=None, key=1, value=2,
                               num_lines=NUM_LINES, no_zero=True):
    """
    .. versionadded:: 0.2.5

    .. versionchanged:: 0.3.0
        added *no_zero*

    This function superseeds :func:`parse_gi_taxa_table`, since NCBI is
    deprecating the GIDs in favor of accessions like *X53318*. The new file can
    be found at the NCBI `ftp://ftp.ncbi.nih.gov/pub/taxonomy/accession2taxid`,
    for DNA sequences (*nt* DB) *nucl_gb.accession2taxid.gz*.

    The file contains 4 columns, the first one is the accession without its
    version, the second one includes the version, the third column is the
    taxonomic identifier and the fourth is either the old GID or **na**.

    The column used as key is the *second*, since by default the fasta headers
    used in NCBI DBs use the versioned identifier. To use the GID as key, the
    *key* parameter can be set to 3, but if no identifier is found (*na* as per
    the file README), the line is skipped.

    Arguments:
        file_handle (str, file): file name or open file handle
        acc_ids (None, list): if it's not `None` only the keys included in the
            passed `acc_ids` list will be returned
        key (int): 0-based index for the column to use as accession. Defaults
            to the versioned accession that is used in GenBank fasta files.
        num_lines (None, int): number of which a message is logged. If None,
            no message is logged
        no_zero (bool): if True (default) a key with taxon_id of 0 is not yield

    .. note::

        GIDs are being phased out in September 2016:
        http://www.ncbi.nlm.nih.gov/news/03-02-2016-phase-out-of-GI-numbers/

    """

    file_handle = open_file(file_handle, 'rb')

    LOG.info(
        "Reading taxonomic information from file (%s)",
        getattr(file_handle, 'name', repr(file_handle))
    )

    if acc_ids is not None:
        acc_ids = set(acc_ids)

    zero_acc = 0

    idx = 0

    for idx, line in enumerate(file_handle):
        line = line.decode('ascii')
        # skip header
        if line.startswith('accession'):
            continue

        if (num_lines is not None) and ((idx + 1) % num_lines == 0):
            LOG.info("Parsed %d lines", idx + 1)

        line = line.strip().split('\t')

        acc_id = line[key]

        if (acc_ids is not None) and (acc_id not in acc_ids):
            continue

        # this is the case if the column used as key is the fourth (GID) and
        # the GID is not available for that accession. The best case is to skip
        if acc_id.lower() == 'na':
            continue

        taxon_id = int(line[value])

        # in cases where 0 is the organism (doesn't exist in the taxonomy)
        if no_zero and (taxon_id == 0):
            # LOG.warning("accession '%s' has taxon_id 0", acc_id)
            zero_acc += 1
            continue

        yield acc_id, taxon_id

    if no_zero and (zero_acc > 0):
        LOG.warning("%d accessions have taxon_id 0", zero_acc)

    if idx == 0:
        LOG.error("The file passed is empty")
    else:
        LOG.info("Parsed %d lines", idx + 1)



import sys
import os
# import ez_setup
# ez_setup.use_setuptools()
import setuptools
from distutils.extension import Extension

try:
    USE_CYTHON = os.environ['USE_CYTHON']
    USE_CYTHON = True
except KeyError:
    USE_CYTHON = False

ext = '.pyx' if USE_CYTHON else '.c'
extensions = [
    Extension("mgkit.utils._sequence", ["mgkit/utils/_sequence" + ext])
]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)

__VERSION__ = "0.5.3"

from setuptools import setup, find_packages

install_requires = [
    'numpy>=1.9.2',
    'pandas>=0.24',
    'tqdm>=4.0',
    'semidbm>=0.5.1',
    'pymongo>=3.1.1',
    'scipy>=0.15.1',
    'matplotlib>=2',
    'msgpack-python>=0.4.6',
    'statsmodels>=0.8',
    'networkx',
    'future',
    'requests',
    'click>=6',
    #support for enum backported from Python 3.4
    'enum34;python_version<"3.4"',
]

# Build of documentation fails on RTD when pytables is
# required
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    install_requires.append('tables>=3.4.2')
    install_requires.append('HTSeq>=0.9.1')
    install_requires.append('pysam>=0.14')


with open('README.rst') as file:
    long_description = file.read()


setup(
    name="mgkit",
    version=__VERSION__,
    packages=find_packages(
        exclude=['tests']
    ),
    #content of readme file to be used on PyPI as home page
    long_description=long_description,
    # package_dir={'mgkit': 'mgkit'},
    install_requires=install_requires,
    scripts=[
        # 'bin/snp_analysis.py',
        'scripts/download-taxonomy.sh',
        'scripts/download-uniprot-taxa.sh',
        'scripts/download-ncbi-taxa.sh',
        'scripts/sort-gff.sh',
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest>=3.5',
        'pytest-datadir',
        'pytest-console-scripts'
    ],
    entry_points={
        'console_scripts': [
            'filter-gff = mgkit.workflow.filter_gff:main',
            'add-gff-info = mgkit.workflow.add_gff_info:main',
            'get-gff-info = mgkit.workflow.extract_gff_info:main',
            'hmmer2gff = mgkit.workflow.hmmer2gff:main',
            'blast2gff = mgkit.workflow.blast2gff:main',
            'snp_parser = mgkit.workflow.snp_parser:main',
            'fastq-utils = mgkit.workflow.fastq_utils:main',
            'taxon-utils = mgkit.workflow.taxon_utils:main',
            'json2gff = mgkit.workflow.json2gff:main',
            'fasta-utils = mgkit.workflow.fasta_utils:main',
            'sampling-utils = mgkit.workflow.sampling_utils:main',
            'edit-gff = mgkit.workflow.edit_gff:main',
            'pnps-gen = mgkit.workflow.pnps_gen:main',
        ],
    },
    author="Francesco Rubino",
    author_email="rubino.francesco@gmail.com",
    description="Metagenomics Framework",
    license="GPL2+",
    keywords="metagenomics library biology bioinformatics snps gff fasta",
    url="https://github.com/frubino/mgkit",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    ext_modules=extensions,
)



"""
Commands
--------

* Interleave/deinterleave paired-end fastq files.
* Converts to FASTA
* sort 2 files to sync the headers

Changes
-------

.. versionchanged:: 0.3.4
    moved to use click, internal fastq parsing, removed *rand* command

.. versionchanged:: 0.3.1
    added stdin/stdout defaults for some commands

.. versionchanged:: 0.3.0
    added *convert* command to FASTA

"""
from __future__ import division
from builtins import zip
import logging
import click
import mgkit
from mgkit.io.fastq import choose_header_type, write_fastq_sequence, load_fastq
from mgkit.io import fasta
from . import utils

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('convert', help="""Convert FastQ file [fastq-file] to FASTA file
              [fasta-file]""")
@click.option('-v', '--verbose', is_flag=True)
@click.argument('fastq-file', type=click.File('rb'), default='-')
@click.argument('fasta-file', type=click.File('wb'), default='-')
def convert_command(verbose, fastq_file, fasta_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        "Writing FASTA file (%s)",
        getattr(fasta_file, 'name', repr(fasta_file))
    )

    for seq_id, seq, qual in load_fastq(fastq_file):
        fasta.write_fasta_sequence(fasta_file, seq_id, seq)


def report_counts(count, wcount, counter=None):
    "Logs the status"
    if (counter is None) or (counter % 100000 == 0):
        LOG.info(
            "Read %-9d sequences, Wrote %-9d (%.2f%%)",
            count,
            wcount,
            wcount / count * 100
        )


@main.command('sort', help="""Sort paired-end sequences from [mate1-input] and
              [mate2-input] into files [mate1-output] and [mate2-output]""")
@click.option('-v', '--verbose', is_flag=True)
@click.argument('mate1-input', type=click.File('rb'))
@click.argument('mate2-input', type=click.File('rb'))
@click.argument('mate1-output', type=click.File('wb'))
@click.argument('mate2-output', type=click.File('wb'))
def sort(verbose, mate1_input, mate2_input, mate1_output, mate2_output):
    "Sort two fastq files"

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing [mate1-output] to file (%s)',
        getattr(mate1_output, 'name', repr(mate1_output))
    )
    LOG.info(
        'Writing [mate2-output] to file (%s)',
        getattr(mate2_output, 'name', repr(mate2_output))
    )

    regex = None
    simple_header = False

    mate1 = {}
    mate2 = {}

    count = 0
    wcount = 0

    for (seq_id1, seq1, qual1), (seq_id2, seq2, qual2) in zip(load_fastq(mate1_input), load_fastq(mate2_input)):

        count += 1

        if (regex is None) and (not simple_header):
            regex = choose_header_type(seq_id1)
            if regex is None:
                simple_header = True
                LOG.info("Using a simple header structure")

        if simple_header:
            key1 = seq_id1[:-1]
            key2 = seq_id2[:-1]
        else:
            match1 = regex.search(seq_id1)
            match2 = regex.search(seq_id2)

            key1 = (
                match1.group('lane'),
                match1.group('tile'),
                match1.group('xcoord'),
                match1.group('ycoord')
            )
            key2 = (
                match2.group('lane'),
                match2.group('tile'),
                match2.group('xcoord'),
                match2.group('ycoord')
            )

        seq1 = (seq_id1, seq1, qual1)
        seq2 = (seq_id2, seq2, qual2)

        if key1 == key2:
            # if the 2
            write_fastq_sequence(mate1_output, *seq1)
            write_fastq_sequence(mate2_output, *seq2)
            wcount += 1
            report_counts(count, wcount, count)
            continue

        mate1[key1] = seq1
        mate2[key2] = seq2

        if key1 in mate2:
            write_fastq_sequence(mate1_output, *mate1[key1])
            write_fastq_sequence(mate2_output, *mate2[key1])
            del mate1[key1]
            del mate2[key1]
            wcount += 1
        if key2 in mate1:
            write_fastq_sequence(mate1_output, *mate1[key2])
            write_fastq_sequence(mate2_output, *mate2[key2])
            del mate1[key2]
            del mate2[key2]
            wcount += 1

        report_counts(count, wcount, count)

    report_counts(count, wcount, None)


@main.command('di', help="""Deinterleave sequences from [fastq-file], into
              [mate1-file] and [mate2-file]""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-s', '--strip', is_flag=True, default=False,
              help="Strip additional info")
@click.argument('fastq-file', type=click.File('rb'), default='-')
@click.argument('mate1-file', type=click.File('wb'))
@click.argument('mate2-file', type=click.File('wb'))
def deinterleave(verbose, strip, fastq_file, mate1_file, mate2_file):
    "Deinterleave a fastq file"

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing [mate1-file] to file (%s)',
        getattr(mate1_file, 'name', repr(mate1_file))
    )
    LOG.info(
        'Writing [mate2-file] to file (%s)',
        getattr(mate2_file, 'name', repr(mate2_file))
    )

    regex = None
    simple_header = False

    mate1 = {}
    mate2 = {}

    count = 0
    wcount = 0

    for seq_id, seq, qual in load_fastq(fastq_file):

        count += 1

        if (regex is None) and (not simple_header):
            regex = choose_header_type(seq_id)
            if regex is None:
                LOG.info("Using a simple header structure")
                simple_header = True

        if simple_header:
            key = seq_id[:-1]
            mate = int(seq_id[-1])
        else:
            match = regex.search(seq_id)
            key = (
                match.group('lane'),
                match.group('tile'),
                match.group('xcoord'),
                match.group('ycoord')
            )
            mate = int(match.group('mate'))

        if strip:
            sequence_name = seq_id.split('\t')[0]
        else:
            sequence_name = seq_id

        if mate == 1:
            mate1[key] = (sequence_name, seq, qual)
        else:
            mate2[key] = (sequence_name, seq, qual)

        try:
            # if sequence header in both
            seq1 = mate1[key]
            seq2 = mate2[key]
            write_fastq_sequence(mate1_file, *seq1)
            write_fastq_sequence(mate2_file, *seq2)
            wcount += 2
            del mate1[key]
            del mate2[key]
        except KeyError:
            pass

        report_counts(count, wcount, count)

    report_counts(count, wcount, None)


@main.command('il', help="""Interleave sequences from [mate1-file] and
              [mate2-file] into [fastq-file]""")
@click.option('-v', '--verbose', is_flag=True)
@click.argument('mate1-file', type=click.File('rb'))
@click.argument('mate2-file', type=click.File('rb'))
@click.argument('fastq-file', type=click.File('wb'), default='-')
def interleave(verbose, mate1_file, mate2_file, fastq_file):
    "Interleave two fastq files"

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing interleaved [fastq-file] to file (%s)',
        getattr(fastq_file, 'name', repr(fastq_file))
    )

    regex = None
    simple_header = False

    mate1 = {}
    mate2 = {}

    count = 0
    wcount = 0

    for (seq_id1, seq1, qual1), (seq_id2, seq2, qual2) in zip(load_fastq(mate1_file), load_fastq(mate2_file)):

        count += 1

        if (regex is None) and (not simple_header):
            regex = choose_header_type(seq_id1)
            if regex is None:
                simple_header = True
                LOG.info("Using a simple header structure")

        if simple_header:
            key1 = seq_id1[:-1]
            key2 = seq_id2[:-1]
        else:
            match1 = regex.search(seq_id1)
            match2 = regex.search(seq_id2)

            key1 = (
                match1.group('lane'),
                match1.group('tile'),
                match1.group('xcoord'),
                match1.group('ycoord')
            )
            key2 = (
                match2.group('lane'),
                match2.group('tile'),
                match2.group('xcoord'),
                match2.group('ycoord')
            )

        seq1 = (seq_id1, seq1, qual1)
        seq2 = (seq_id2, seq2, qual2)

        if key1 == key2:
            # if the 2
            write_fastq_sequence(fastq_file, *seq1)
            write_fastq_sequence(fastq_file, *seq2)
            wcount += 1
            report_counts(count, wcount, wcount)
            continue

        mate1[key1] = seq1
        mate2[key2] = seq2

        if key1 in mate2:
            write_fastq_sequence(fastq_file, *mate1[key1])
            write_fastq_sequence(fastq_file, *mate2[key1])
            del mate1[key1]
            del mate2[key1]
            wcount += 1
        if key2 in mate1:
            write_fastq_sequence(fastq_file, *mate1[key2])
            write_fastq_sequence(fastq_file, *mate2[key2])
            del mate1[key2]
            del mate2[key2]
            wcount += 1

        report_counts(count, wcount, count)

    report_counts(count, wcount, None)



"""
Utilities to map genes
"""

from builtins import zip
import pandas
import itertools


def count_genes_in_mapping(gene_lists, labels, mapping, normalise=False):
    """
    Maps lists of ids to a mapping dictionary, returning a
    :class:`pandas.DataFrame` in which the rows are the labels provided and
    the columns the categories to which the ids map. Each element of the matrix
    label-category is the sum of all ids in the relative gene list that maps to
    the specific category.

    :param iterable gene_lists: an iterable in which each element is a iterable
        of ids that can be mapped to mapping
    :param iterable labels: an iterable of strings that defines the labels to
        be used in the resulting rows in the :class:`pandas.DataFrame`; must
        have the same length as gene_lists
    :param dict mapping: a dictionary in the form:
        gene_id->[cat1, cat2, .., catN]
    :param bool normalise: if True the counts are normalised over the total for
        each row.

    :return: a :class:`pandas.DataFrame` instance
    """
    categories = set(itertools.chain(*list(mapping.values())))

    matrix = pandas.DataFrame(index=labels, columns=categories)
    matrix.fillna(0, inplace=True)

    for label, gene_list in zip(labels, gene_lists):
        for gene_id in gene_list:
            try:
                mapped_ids = mapping[gene_id]
            except KeyError:
                continue

            for mapped_id in mapped_ids:
                matrix.ix[label, mapped_id] += 1

    matrix = matrix.ix[:, matrix.sum() > 0]

    if normalise:
        matrix = matrix.div(matrix.sum(axis=1), axis=0)

    return matrix


def group_annotation_by_mapping(annotations, mapping, attr='ko'):
    """
    Group annotations by mapping dictionary

    :param iterable annotations: iterable of :class:`gff.GFFKeg` instances
    :param dict mapping: dictionary with mappings for the attribute requested
    :param str attr: attribute of the annotation to be used as key in mapping

    :return dict: dictionary category->annotations
    """
    grouped = dict(
        (categ, []) for categ in set(itertools.chain(*list(mapping.values())))
    )
    for annotation in annotations:
        try:
            categories = mapping[getattr(annotation.attributes, attr)]
        except KeyError:
            # not included in the categories
            continue
        for categ in categories:
            grouped[categ].append(annotation)

    return grouped



"""

.. versionchanged:: 0.3.4
    using *click* instead of *argparse*

.. versionadded:: 0.2.6

This script converts annotations in JSON format that were created using MGKit
back into GFF annotations.

mongodb command
***************

Annotations converted into MongoDB records with *get-gff-info mongodb* can be
converted back into a GFF file using this command. It can be useful to get a
GFF file as output from a query to a MongoDB instance on the command line.

For example:

mongoexport -d db -c test | json2gff mongodb

will convert all the annotations in the database *db*, collection *test* to
the standard out.

"""

from __future__ import division
import click
import logging
import json

import mgkit
from . import utils
from mgkit.io import gff

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('mongodb', help="""Convert annotations from a MongoDB instance to
              GFF""")
@click.option('-v', '--verbose', is_flag=True)
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('gff-file', type=click.File('wb'), default='-')
def mongodb_command(verbose, input_file, gff_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        'Writing to file (%s)',
        getattr(gff_file, 'name', repr(gff_file))
    )

    LOG.info(
        'Reading from file (%s)',
        getattr(input_file, 'name', repr(input_file))
    )

    for line in input_file:
        line = line.decode('ascii')
        annotation = gff.from_mongodb(json.loads(line), lineage=False)
        annotation.to_file(gff_file)



"""
.. versionadded:: 0.1.14

Contains code related to colour
"""
import logging


LOG = logging.getLogger(__name__)


def float_to_hex_color(r, g, b):
    """
    .. versionadded:: 0.1.14

    Converts RGB float values to Hexadecimal value string
    """
    convert = lambda x: int(x * 255)

    return "#{0:02x}{1:02x}{2:02x}".format(convert(r), convert(g), convert(b))


def palette_float_to_hex(palette):
    """
    .. versionadded:: 0.1.16

    Applies :func:`float_to_hex_color` to an iterable of colors
    """

    colors = []

    for color in palette:
        if len(color) > 3:
            color = color[:3]
        colors.append(float_to_hex_color(*color))

    return colors

__all__ = ['float_to_hex_color', 'palette_float_to_hex']



"""
Filters GFF annotations in different ways.

Value Filtering
***************

Enables filtering of GFF annotations based on the the values of attributes of a
GFF annotation. The filters are based on equality of numbers (internally
converted into float) and strings, a string contained in the value of an attribute
less or greater than are included as well. The length of annotation has the
attribute *length* and can be tested.

Overlap Filtering
*****************

Filters overlapping annotations using the functions
:func:`mgkit.filter.gff.choose_annotation` and
:func:`mgkit.filter.gff.filter_annotations`, after the annotations are grouped
by both sequence and strand. If the GFF is sorted by sequence name and strand,
the `-t` can be used to make the filtering use less memory. It can be sorted in
Unix using `sort -s -k 1,1 -k 7,7 gff_file`, which applies a stable sort using
the sequence name as the first key and the strand as the second key.

.. note::

    It is also recommended to use::

        export LC_ALL=C

    To speed up the sorting

.. blockdiag::

    {
        orientation = portrait;

        class mgkit [color = "#e41a1c" , textcolor = 'white', width=200,
        fontsize=15];
        class data [color = "#4daf4a" , textcolor = 'white', width=200,
        fontsize=15];
        class software [color = "#377eb8", textcolor = "white", fontsize=15];

        sort [class = software];
        "GFF" [class = data, shape = flowchart.input];
        parse_gff [class = "mgkit"];
        group_annotations [class = "mgkit"];
        filter_annotations [class = "mgkit", shape = flowchart.condition,
        fontsize=12];

        "GFF" -> parse_gff -> sort -> filter_annotations;
        parse_gff -> group_annotations -> filter_annotations;

        "Filtered Annotations" [class = data, stacked];
        filter_annotations -> "Filtered Annotations";
    }

The above digram describes the internals of the script.

The annotations needs first to be grouped by seq_id and strand, forming a group
that can be then be passed to :func:`mgkit.filter.gff.filter_annotations`.
This function:

    #. sort annotations by bit score, from the highest to the lowest
    #. loop over all combination of N=2 annotations:

        #. choose which of the two annotations to discard if they overlap for a
           the required amount of bp (defaults to 100bp)
        #. in which case, the preference is given to the db quality first, than
           the bit score and finally the lenght of annotation, the one with the
           highest values is kept

While the default behaviour is the same, now it is posible to decided the
function used to discard one the two annotations. It is possible to use the
`-c` argument to pass a string that defines the function. The string passed must
start with or without a **+**. Using **+** translates into the builtin function
*max* while no **+** translates into *min* from the second character on, any
number of attributes can be used, separated by commas. The attributes, however,
must be one of the properties defined in :class:`mgkit.io.gff.Annotation`,
*bitscore* that returns the value converted in a *float*. Internally the
attributes are stored as strings, so for attributes that have no properties in
the class, such as *evalue*, the `float` builtin is applied.

The tuples built for both annotations are then passed to the comparison
function to be selected and the value returned by it is **discarded**. The
order of the elements in the string is important to define the priority
given to each element in the comparison and the leftmost one has the
highesst priority.

Examples of function strings:

* `-dbq,bitscore,length` becomes max((ann1.dbq, ann1.bitscore, ann1.length),
  (ann2.dbq, ann2.bitscore, ann2.length) - This is default and previously
  only choice
* `-bitscore,length,dbq` uses the same elements but gives lowest priority
  to *dbq*
* `+evalue`: will discard the annotation with the highest *evalue*

Per Sequence Values
*******************

The *sequence* command allows to filter on a per sequence basis, using
functions such as the median, quantile and mean on attributes like evalue,
bitscore and identity. The file can be passed as sorted already, saving memory
(like in the *overlap* command), but it's not needed to sort the file by strand,
only by the first column.

Coverage Filtering
******************

The *cov* command calculates the coverage of annotations as a measure of the
percentage of each reference sequence length. A minimum coverage percentage can
be used to keep the annotations of sequences that have a greater or equal
coverage than the specified one.

Changes
*******

.. versionadded:: 0.1.12

.. versionchanged:: 0.1.13
    added *--sorted* option

.. versionchanged:: 0.2.0
    changed option *-c* to accept a string to filter overlap

.. versionchanged:: 0.2.5
    added *sequence* command

.. versionchanged:: 0.2.6
    added *length* as attribute and *min*/*max*, and *ge* is the default
    comparison for command *sequence*, *--sort-attr* to *overlap*

.. versionchanged:: 0.3.1
    added *--num-gt* and *--num-lt* to *values* command, added *cov* command

.. versionchanged:: 0.3.4
    moved to use *click* for argument parsing reworked the *values*, *sequence*
    commands

.. versionchanged:: 0.4.4
    *overlap* command: added option to not use the strand information and
    added an option to make multiple passes of overlap for each sequence

"""
from __future__ import division
from future.utils import viewvalues
import logging
import functools
import click
import pandas
from tqdm import tqdm
import mgkit
from . import utils
from ..io import gff, fasta
from ..filter import gff as filter_gff
from ..utils.common import ranges_length

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


def perseq_calc_threshold(annotations, attribute, function, func_arg=None):

    values = pandas.Series(
        annotation.get_attr(attribute, float)
        for annotation in annotations
    )

    if function == 'mean':
        thres = values.mean()
    elif function == 'median':
        thres = values.median()
    elif function == 'quantile':
        thres = values.quantile(func_arg)
    elif function == 'std':
        thres = values.mean() + (func_arg * values.std())
    elif function == 'max':
        thres = values.max()
    elif function == 'min':
        thres = values.min()

    LOG.debug(
        "Threshold for contig %s found: %.2f, using funcion %s and its " +
        "arg %.2f on %d ann.",
        annotations[0],
        thres,
        function,
        func_arg if function in ('quantile', 'std') else 0,
        len(annotations)
    )
    return thres


def find_comparison(comparison):
    if comparison == 'gt':
        return lambda x, y: x > y
    elif comparison == 'ge':
        return lambda x, y: x >= y
    elif comparison == 'lt':
        return lambda x, y: x < y
    elif comparison == 'le':
        return lambda x, y: x <= y


@main.command('sequence', help='Filter on a per sequence basis')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--sorted', default=False, is_flag=True,
              help='''If the GFF file is sorted (all of a sequence annotations
              are contiguos) can use less memory, `sort -s -k 1,1` can be used''',)
@click.option('-a', '--attribute', default='bitscore', show_default=True,
              type=click.Choice(['evalue', 'bitscore', 'identity', 'length']),
              help='Attribute on which to apply the filter')
@click.option('-f', '--function', show_default=True,
              type=click.Choice(['mean', 'median', 'quantile', 'std', 'max', 'min']),
              default='mean', help='Function for filtering')
@click.option('-l', '--value', default=None, type=click.FLOAT,
              help='Value for the function (used for *std* and *quantile*)')
@click.option('-c', '--comparison', default='ge', show_default=True,
              type=click.Choice(['gt', 'ge', 'lt', 'le']),
              help='Type of comparison (e.g. ge -> greater than or equal to)')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def filter_perseq(verbose, sorted, attribute, function, value, comparison,
                  progress, input_file, output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if (function in ['std', 'quantile']) and value is None:
        utils.exit_script('When *std* or *quantile* are used, --value must be passed', 1)

    comparison = find_comparison(comparison)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    file_iterator = gff.parse_gff(input_file, gff_type=gff.from_gff)

    if sorted:
        LOG.info("Input GFF is assumed sorted")
        grouped = gff.group_annotations_sorted(
            file_iterator, lambda x: x.seq_id
        )
    else:
        grouped = viewvalues(gff.group_annotations(
            file_iterator, lambda x: x.seq_id
        ))

    if progress:
        grouped = tqdm(grouped)

    for annotations in grouped:
        threshold = perseq_calc_threshold(
            annotations,
            attribute,
            function,
            value
        )

        for annotation in annotations:
            if comparison(annotation.get_attr(attribute, float), threshold):
                annotation.to_file(output_file)


@main.command('cov', help='Filter on a per coverage basis')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-f', '--reference', type=click.File('rb'), required=True,
              help='Reference FASTA file for the GFF')
@click.option('-s', '--strand-specific', default=False, is_flag=True,
              help='If the coverage must be calculated on each strand')
@click.option('-t', '--sorted', default=False, is_flag=True,
              help='Assumes the GFF to be correctly sorted')
@click.option('-c', '--min-coverage', default=0., type=click.FLOAT,
              help='Minimum coverage for the contig/strand')
@click.option('-r', '--rename', default=False, is_flag=True,
              help='Emulates BLAST in reading the FASTA file (keeps only the header before the first space)')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def coverage_command(verbose, reference, strand_specific, sorted, min_coverage,
                     rename, progress, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    file_iterator = gff.parse_gff(input_file, gff_type=gff.from_gff)

    if strand_specific:
        def key_func(x): return (x.seq_id, x.strand)
    else:
        def key_func(x): return x.seq_id

    if sorted:
        LOG.info("Input GFF is assumed sorted")
        grouped = gff.group_annotations_sorted(
            file_iterator, key_func
        )
    else:
        grouped = viewvalues(gff.group_annotations(
            file_iterator, key_func
        ))

    seq_iter = fasta.load_fasta_rename(reference) if rename else fasta.load_fasta(reference)

    sequences = {
        seq_id: len(seq)
        for seq_id, seq in seq_iter
    }

    if progress:
        grouped = tqdm(grouped)

    for annotations in grouped:
        covered = ranges_length(
            gff.elongate_annotations(annotations)
        ) / sequences[annotations[0].seq_id] * 100

        if covered < min_coverage:
            continue

        for annotation in annotations:
            annotation.to_file(output_file)


def filter_eq(annotation, attr=None, value=None, conv=None):
    try:
        ann_value = annotation.get_attr(attr, conv=conv)
    except gff.AttributeNotFound:
        return False
    return ann_value == value


def filter_in(annotation, attr=None, value=None, conv=None):
    try:
        ann_value = annotation.get_attr(attr, conv=conv)
    except gff.AttributeNotFound:
        return False
    return value in ann_value


def filter_gt(annotation, attr=None, value=None, conv=None, equal=None):
    try:
        ann_value = annotation.get_attr(attr, conv=conv)
    except gff.AttributeNotFound:
        return False
    if equal:
        return ann_value >= value
    else:
        return ann_value > value


def filter_lt(annotation, attr=None, value=None, conv=None, equal=None):
    try:
        ann_value = annotation.get_attr(attr, conv=conv)
    except gff.AttributeNotFound:
        return False
    if equal:
        return ann_value <= value
    else:
        return ann_value < value


def setup_filters(str_eq, str_in, num_eq, num_ge, num_le, num_gt, num_lt):
    filters = []

    for attr, value in str_eq:
        LOG.info('Filter string (%s = %s)', attr, value)
        filters.append(
            functools.partial(filter_eq, attr=attr, value=value, conv=str)
        )
    for attr, value in str_in:
        LOG.info('Filter string (%s in %s)', value, attr)
        filters.append(
            functools.partial(filter_in, attr=attr, value=value, conv=str)
        )
    for attr, value in num_eq:
        LOG.info('Filter number (%s = %s)', attr, value)
        filters.append(
            functools.partial(filter_eq, attr=attr, value=value, conv=float)
        )
    for attr, value in num_ge:
        LOG.info('Filter number (%s >= %s)', attr, value)
        filters.append(
            functools.partial(filter_gt, attr=attr, value=value, conv=float, equal=True)
        )
    for attr, value in num_gt:
        LOG.info('Filter number (%s > %s)', attr, value)
        filters.append(
            functools.partial(filter_gt, attr=attr, value=value, conv=float, equal=False)
        )
    for attr, value in num_le:
        LOG.info('Filter number (%s <= %s)', attr, value)
        filters.append(
            functools.partial(filter_lt, attr=attr, value=value, conv=float, equal=True)
        )
    for attr, value in num_lt:
        LOG.info('Filter number (%s < %s)', attr, value)
        filters.append(
            functools.partial(filter_lt, attr=attr, value=value, conv=float, equal=False)
        )
    return filters


def validate_params(ctx, param, values, convert=str):
    new_values = []
    for value in values:
        value = value.split(':')
        if len(value) != 2:
            raise click.BadParameter(
                "Wrong key/value format, must be 'key:value' 'key:value'"
            )
        new_values.append((value[0], convert(value[1])))
    return new_values


@main.command('values', help="""Filter based on values""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('--str-eq', multiple=True,
              callback=functools.partial(validate_params, convert=str),
              help='''filter by custom key:value, if the argument is 'key:value' the annotation is kept if it contains an attribute 'key' whose value is exactly 'value' as a string''')
@click.option('--str-in', multiple=True,
              callback=functools.partial(validate_params, convert=str),
              help="Same as '--str-eq' but 'value' is contained in the attribute")
@click.option('--num-eq', multiple=True,
              callback=functools.partial(validate_params, convert=float),
              help="Same as '--str-eq' but 'value' is a number which is equal or greater than")
@click.option('--num-ge', multiple=True,
              callback=functools.partial(validate_params, convert=float),
              help="Same as '--str-eq' but 'value' is a number which is equal or greater than")
@click.option('--num-le', multiple=True,
              callback=functools.partial(validate_params, convert=float),
              help="Same as '--num-ge' but 'value' is a number which is equal or less than")
@click.option('--num-gt', multiple=True,
              callback=functools.partial(validate_params, convert=float),
              help="Same as '--str-eq' but 'value' is a number which is greater than")
@click.option('--num-lt', multiple=True,
              callback=functools.partial(validate_params, convert=float),
              help="Same as '--num-ge' but 'value' is a number which is less than")
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def filter_values(verbose, str_eq, str_in, num_eq, num_ge, num_le, num_gt, num_lt,
                  progress, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    filters = setup_filters(str_eq, str_in, num_eq, num_ge, num_le, num_gt, num_lt)

    iterator = gff.parse_gff(input_file, gff_type=gff.from_gff)
    if progress:
        iterator = tqdm(iterator)

    for annotation in iterator:
        if all(filter_func(annotation) for filter_func in filters):
            annotation.to_file(output_file)


def make_choose_func(values):
    """Builds the function used to choose between two annotations."""
    values = values.strip()

    LOG.info("Filter function used for overlaps %s", values)

    if values.startswith('+'):
        function = max
        values = values[1:]
    else:
        function = min

    attributes = values.split(',')

    def choose_func(a1, a2):
        return function(
            a1,
            a2,
            key=lambda el: tuple(
                getattr(el, attribute, None) if hasattr(el, attribute) else el.get_attr(attribute, float)
                for attribute in attributes
            )
        )

    return choose_func


@main.command('overlap', help='Use overlapping filter')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-s', '--size', type=click.INT, default=100, show_default=True,
              help='Size of the overlap that triggers the filter')
@click.option('-t', '--sorted', is_flag=True, default=False,
              help='''If the GFF file is sorted (all of a sequence annotations are contiguos and sorted by strand) can use less memory, `sort -s -k 1,1 -k 7,7` can be used''')
@click.option('-c', '--choose-func', default='dbq,bitscore,length',
              help='Function to choose between two overlapping annotations')
@click.option('-a', '--sort-attr', default='bitscore', show_default=True,
              type=click.Choice(['bitscore', 'identity', 'length']),
              help='Attribute to sort annotations before filtering (default bitscore)')
@click.option('-d', '--no-strand', default=False, is_flag=True,
              help="Strand information is not used, if '-t' is used, sort GFF file with `sort -s -k 1,1`")
@click.option('-n', '--iterations', type=click.IntRange(min=1, max=10),
              default=1,
              help="Max number of iteration over which filter the overlaps")
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def filter_overlaps(verbose, size, sorted, choose_func, sort_attr, no_strand,
                    iterations, progress, input_file, output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    choose_func = make_choose_func(choose_func)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    LOG.info("Max number of iterations: %d", iterations)

    file_iterator = gff.parse_gff(input_file, gff_type=gff.from_gff)

    if no_strand:
        LOG.info("Annotations strand will be ignored")
        def group_key_func(a): return a.seq_id
    else:
        def group_key_func(a): return (a.seq_id, a.strand)

    if sorted:
        LOG.info("Input GFF is assumed sorted")
        grouped = gff.group_annotations_sorted(file_iterator,
                                               key_func=group_key_func)
    else:
        grouped = viewvalues(gff.group_annotations(file_iterator,
                             key_func=group_key_func))

    choose_func = functools.partial(
        filter_gff.choose_annotation,
        overlap=size,
        choose_func=choose_func
    )

    if progress:
        grouped = tqdm(grouped)

    for annotations in grouped:
        prev_filtered = set(annotations)
        for num_iteration in range(iterations):
            filtered = filter_gff.filter_annotations(
                prev_filtered,
                choose_func=choose_func,
                sort_func=lambda x: x.get_attr(sort_attr, float),
                reverse=True
            )
            # if the number of annotations is the same, there's no need to
            # keep going
            if prev_filtered == filtered:
                # LOG.debug("Number of iterations %d", num_iteration + 1)
                break
            else:
                prev_filtered = filtered

        for annotation in filtered:
            annotation.to_file(output_file)



import pytest
from conftest import skip_no_connection
from mgkit.snps.classes import GeneSNP
from mgkit.snps.filter import filter_genesyn_by_coverage, \
    filter_genesyn_by_gene_id, filter_genesyn_by_taxon_id, FilterFails
import mgkit.snps.mapper
import mgkit.taxon
import functools


@skip_no_connection
def test_snps_taxon1(ncbi_taxonomy):
    # will find it
    filter_list = [
        taxon.taxon_id
        for taxon in ncbi_taxonomy
        if 'methanobrevibacter' in taxon.s_name.lower()
    ]
    taxon_id = ncbi_taxonomy.find_by_name('methanobrevibacter')[0]
    gene_syn = GeneSNP(taxon_id=taxon_id)


    assert filter_genesyn_by_taxon_id(
        gene_syn,
        ncbi_taxonomy,
        filter_list=filter_list,
        exclude=False
    )


@skip_no_connection
def test_snps_taxon1_rev(ncbi_taxonomy):
    # will find it
    filter_list = [
        taxon.taxon_id
        for taxon in ncbi_taxonomy
        if 'methanobrevibacter' in taxon.s_name.lower()
    ]
    taxon_id = ncbi_taxonomy.find_by_name('methanobrevibacter')[0]
    gene_syn = GeneSNP(taxon_id=taxon_id)

    assert not filter_genesyn_by_taxon_id(
        gene_syn,
        ncbi_taxonomy,
        filter_list=filter_list,
        exclude=True
    )


@skip_no_connection
def test_snps_taxon2(ncbi_taxonomy):
    # will find it
    filter_list = [
        taxon.taxon_id
        for taxon in ncbi_taxonomy
        if 'clostridium' in taxon.s_name
    ]
    taxon_id = ncbi_taxonomy.find_by_name('methanobrevibacter')[0]
    gene_syn = GeneSNP(taxon_id=taxon_id)

    assert not filter_genesyn_by_taxon_id(
        gene_syn,
        ncbi_taxonomy,
        filter_list=filter_list,
        exclude=False
    )


@skip_no_connection
def test_snps_taxon2_rev(ncbi_taxonomy):
    # will find it
    filter_list = [
        taxon.taxon_id
        for taxon in ncbi_taxonomy
        if 'clostridium' in taxon.s_name
    ]
    taxon_id = ncbi_taxonomy.find_by_name('methanobrevibacter')[0]
    gene_syn = GeneSNP(taxon_id=taxon_id)

    assert filter_genesyn_by_taxon_id(
        gene_syn,
        ncbi_taxonomy,
        filter_list=filter_list,
        exclude=True
    )


@skip_no_connection
def test_snps_taxon3(ncbi_taxonomy):
    # will find it
    filter_list = [
        taxon.taxon_id
        for taxon in ncbi_taxonomy
        if 'clostridium' in taxon.s_name
    ]
    taxon_id = ncbi_taxonomy.find_by_name('methanobrevibacter')[0]
    gene_syn = GeneSNP(taxon_id=taxon_id)

    assert not filter_genesyn_by_taxon_id(
        gene_syn,
        ncbi_taxonomy,
        filter_list=filter_list,
        exclude=False,
        func=mgkit.taxon.is_ancestor
    )


def test_snps_taxon4_exc1():
    gene_syn = GeneSNP()
    with pytest.raises(FilterFails):
        filter_genesyn_by_taxon_id(
            gene_syn, filter_list=range(10), taxonomy=None, func=filter
        )


def test_snps_taxon4_exc2():
    gene_syn = GeneSNP()
    with pytest.raises(FilterFails):
        filter_genesyn_by_taxon_id(
            gene_syn, filter_list=None
        )


def test_snps_gene_id1():
    gene_syn = GeneSNP(gene_id='K01201')
    gene_list = ['K01201', 'K02201', 'K01251']

    assert filter_genesyn_by_gene_id(
        gene_syn,
        gene_list,
        id_func=lambda x: x.gene_id
    )


def test_snps_gene_id2():
    gene_syn = GeneSNP(gene_id='K01201')
    gene_list = ['K01201', 'K02201', 'K01251']

    assert not filter_genesyn_by_gene_id(
        gene_syn,
        gene_list,
        exclude=True,
        id_func=lambda x: x.gene_id
    )


def test_snps_gene_id_exc():
    gene_syn = GeneSNP(gene_id='K01201')
    gene_list = None
    with pytest.raises(FilterFails):
        filter_genesyn_by_gene_id(gene_syn, gene_ids=gene_list)


def test_snps_gene_coverage1():
    gene_syn = GeneSNP(gene_id='K01201', coverage=4)
    min_cov = 4

    assert filter_genesyn_by_coverage(gene_syn, min_cov=min_cov)


def test_snps_gene_coverage2():
    gene_syn = GeneSNP(gene_id='K01201', coverage=3)
    min_cov = 4

    assert not filter_genesyn_by_coverage(gene_syn, min_cov=min_cov)


def test_snps_gene_coverage_exc():
    gene_syn = GeneSNP(gene_id='K01201')
    min_cov = None
    with pytest.raises(FilterFails):
        filter_genesyn_by_coverage(gene_syn, min_cov=min_cov)


def test_map_gene_id1():
    gene_map = {
        'K1': ['K2', 'K3']
    }
    assert list(mgkit.snps.mapper.map_gene_id('K1', gene_map)) == gene_map['K1']


def test_map_gene_id2():
    gene_map = {
        'K1': ['K2', 'K3']
    }

    assert list(mgkit.snps.mapper.map_gene_id('K2', gene_map)) == []


@skip_no_connection
def test_map_taxon_id_to_rank1(ncbi_taxonomy):
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_rank(
            839,
            rank='genus',
            taxonomy=ncbi_taxonomy,
            include_higher=False
        )
    ) == [838]


@skip_no_connection
def test_map_taxon_id_to_rank2(ncbi_taxonomy):
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_rank(
            2,
            rank='genus',
            taxonomy=ncbi_taxonomy,
            include_higher=False
        )
    ) == []


@skip_no_connection
def test_map_taxon_id_to_rank3(ncbi_taxonomy):
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_rank(
            2,
            rank='genus',
            taxonomy=ncbi_taxonomy,
            include_higher=True
        )
    ) == [2]


@skip_no_connection
def test_map_taxon_id_to_ancestor1(ncbi_taxonomy):
    func = functools.partial(
        mgkit.taxon.is_ancestor,
        ncbi_taxonomy
    )
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_ancestor(
            839,
            anc_ids=[838],
            func=func
        )
    ) == [838]


@skip_no_connection
def test_map_taxon_id_to_ancestor2(ncbi_taxonomy):
    func = functools.partial(
        mgkit.taxon.is_ancestor,
        ncbi_taxonomy
    )
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_ancestor(
            839,
            anc_ids=[838, 2],
            func=func
        )
    ) == [838, 2]


@skip_no_connection
def test_map_taxon_id_to_ancestor3(ncbi_taxonomy):
    func = functools.partial(
        mgkit.taxon.is_ancestor,
        ncbi_taxonomy
    )
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_ancestor(
            839,
            anc_ids=[838, 2, 1485],
            func=func
        )
    ) == [838, 2]


@skip_no_connection
def test_map_taxon_id_to_ancestor4(ncbi_taxonomy):
    func = functools.partial(
        mgkit.taxon.is_ancestor,
        ncbi_taxonomy
    )
    assert list(
        mgkit.snps.mapper.map_taxon_id_to_ancestor(
            839,
            anc_ids=[2172, 1485],
            func=func
        )
    ) == []
